# Copyright 2021  Facebook. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICEnSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" bASIS,
# wIthOUt wARRAntIES OR COnDItIOnS OF AnY KInD, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# This code is modified by Zilliz.


import torch
from torch import nn
from typing import List, Tuple


class AttentionPool(nn.Module):
    """
    A MLP block that contains two linear layers with a normalization layer.
    the MLP block is used in a transformer model after the attention block.
    ::
                                        Input
                                           ↓
                                        Reshape
                                           ↓
                                          Pool
                                           ↓
                                        Reshape
                                           ↓
                                          norm
    Args:
        thw_shape(List):
            the shape of the input tensor (before flattening).
        pool(Callable):
            Pool operation that is applied to the input tensor.
            If pool is None, return the input tensor.
        has_cls_embed(bool):
            whether the input tensor contains cls token. Pool operation excludes cls token.
        norm(Callable):
            Optional normalization operation applied to tensor after pool.
    Returns:
        tensor(torch.Tensor):
            Input tensor after pool.
        thw_shape(List[int]):
            Output tensor shape (before flattening).
    """
    def __init__(
        self,
        thw_shape,
        pool=None,
        has_cls_embed=True,
        norm=None
    ) -> None:
        super().__init__()
        self.pool = pool
        self.thw_shape = thw_shape
        self.has_cls_embed = has_cls_embed
        self.norm = norm

    def forward(self, x: torch.Tensor) -> Tuple[torch.Tensor, List[int]]:
        if self.pool is None:
            return x, self.thw_shape
        tensor_dim = x.ndim
        if tensor_dim == 4:
            pass
        elif tensor_dim == 3:
            x = x.unsqueeze(1)
        else:
            raise NotImplementedError(f"Unsupported input dimension {x.shape}")

        if self.has_cls_embed:
            cls_tok, x = x[:, :, :1, :], x[:, :, 1:, :]

        b, n, _, c = x.shape
        t, h, w = self.thw_shape
        x = x.reshape(b * n, t, h, w, c).permute(0, 4, 1, 2, 3).contiguous()

        x = self.pool(x)

        thw_shape = [x.shape[2], x.shape[3], x.shape[4]]
        l_pooled = x.shape[2] * x.shape[3] * x.shape[4]
        x = x.reshape(b, n, c, l_pooled).transpose(2, 3)
        if self.has_cls_embed:
            x = torch.cat((cls_tok, x), dim=2)
        if self.norm is not None:
            x = self.norm(x)

        if tensor_dim == 4:
            pass
        else:  # For the case tensor_dim == 3.
            x = x.squeeze(1)
        return x, thw_shape



# Original pytorch implementation by:
# 'An Image Is Worth 16 x 16 Words: Transformers for Image Recognition at Scale'
#       - https://arxiv.org/abs/2010.11929
# 'How to train your ViT? Data, Augmentation, and Regularization in Vision Transformers'
#       - https://arxiv.org/abs/2106.10270
#
# Model weights are provided by corresponding links and credit to original authors.
# Original code by / Copyright 2020, Ross Wightman.
# Modifications & additions by / Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
from torch.utils import model_zoo


def load_pretrained_model(model, configs: dict, weights_path: str = None):
    if weights_path:
        state_dict = torch.load(weights_path)
    else:
        state_dict = model_zoo.load_url(configs['url'], map_location=torch.device('cpu'))
    res = model.load_state_dict(state_dict, strict=False)
    return res

def base_configs():
    return dict(
        img_size=224,
        patch_size=16,
        num_classes=1000,
        embed_dim=768,
        depth=12,
        num_heads=12,
        drop_ratio=0,
    )


def get_configs(model_name):
    if model_name == 'vit_base_patch16_224':
        configs = base_configs()
        configs.update(dict(
            url='https://github.com/jaelgu/towhee_test/releases/download/v1.0.0/vit_base_patch16_224.pth'
        ))
        return configs



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Optional


FRAME = 'frame'


class _Frame:
    """
    Row info of dataframe.
    row_id (`int`):
        Row id of dataframe, an auto-increment data.
    parent_path (`str`):
        FlatMap's output data, will set the data.

        Example:
        The parent_path changes between flatmap operators.

                   dataframe                  dataframe                    dataframe

                    frame                       frame                       frame
                 ┌───────────┐               ┌───────────┐               ┌───────────┐
           f1    │1  '' None │     f1-1      │1   1  None│       f1-1-1  │1  1-1 None│
                 ├───────────┤               ├───────────┤               ├───────────┤
           f2    │2  '' None │     f1-2      │2   1  None│       f1-1-2  │2  1-1 None│
                 ├───────────┤               ├───────────┤               ├───────────┤
                 │           │     f2-3      │3   2  None│       f1-2-3  │3  1-2 None│
                 ├───────────┤               ├───────────┤               ├───────────┤
                 │           │               │           │       f2-3-4  │4  2-3 None│
                 │           │               │           │               │           │
                 └───────────┘               └───────────┘               └───────────┘
                                 ┌───────────┐           ┌───────────┐
                                 │           │           │           │
                ───────────────► │  FlatMap  ├──────────►│  FlatMap  ├────────────────►
                                 │           │           │           │
                                 └───────────┘           └───────────┘


    timestamp (`int`):
        Video images will have timestamp attribute, the video-decoder will add the info to frames.
    empty(`bool`):
        Filter-operator will set empty = True is this frame is filtered.
    prev_id: (`int`):
        Operator input-data's row_id.
    """

    def __init__(self, row_id: int = -1, parent_path: str = '',
                 timestamp: int = Optional[None], empty: bool = False,
                 prev_id=-1):
        self._row_id = row_id
        self._timestamp = timestamp
        self._parent_path = parent_path
        self._empty = empty
        self._prev_id = prev_id

    @property
    def row_id(self):
        return self._row_id

    @row_id.setter
    def row_id(self, row_id: int):
        self._row_id = row_id

    @property
    def timestamp(self):
        return self._timestamp

    @timestamp.setter
    def timestamp(self, timestamp: int):
        self._timestamp = timestamp

    @property
    def parent_path(self):
        return self._parent_path

    @parent_path.setter
    def parent_path(self, parent_path: str):
        self._parent_path = parent_path

    @property
    def prev_id(self):
        return self._prev_id

    @prev_id.setter
    def prev_id(self, prev_id: int):
        self._prev_id = prev_id

    @property
    def empty(self):
        return self._empty

    @empty.setter
    def empty(self, empty: bool):
        self._empty = empty

    def __str__(self) -> str:
        return f'row_id={self.row_id}:parent_path={self.parent_path}:timestamp={self.timestamp}:empty={self.empty}:prev_id={self.prev_id}'



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np


class Image(np.ndarray):
    """
    This class represents an image object. The image data is a numpy.ndarray.

    Agrs:
        mode (`str`):
            The mode of the image(i.e. 'RGB', 'BGR', 'RGBA', 'HSV', etc.).
    """

    def __new__(cls, data: np.ndarray, mode: str = None):
        # Cast `np.ndarray` to be `Image`.
        # See https://numpy.org/doc/stable/user/basics.subclassing.html for details.
        obj = np.asarray(data).view(cls)
        obj._mode = mode
        return obj

    def __array_finalize__(self, obj):
        # `self` is a new object resulting from super().__new__(cls, ...), therefore it
        # only has attributes that the ndarray.__new__ constructor gave it -
        # i.e. those of a standard ndarray.
        if obj is None:
            return
        self._mode = getattr(obj, '_mode', None)

    def __str__(self):
        return 'Image' + ' shape=' + repr(self.shape) + ' mode=' + self.mode

    @property
    def mode(self):
        return self._mode

    def cv2_to_rgb(self) -> np.ndarray:
        # pylint: disable=import-outside-toplevel
        """
        Convert the OpenCV images from BGR to RGB.
        """
        from towhee.utils.cv2_utils import cv2
        return cv2.cvtColor(self, cv2.COLOR_BGR2RGB)



# original code from https://github.com/krasserm/perceiver-io
# modified by Zilliz


from torch import nn


class MultiHeadAttention(nn.Module):
    """
    Multi head attention for Perceiver https://arxiv.org/pdf/2103.03206.pdf.
    Args:
        num_q_channels (`int`):
            Number of q channels.
        num_kv_channels (`int`):
            Number of k or v channels. k has the same channels as v.
        num_heads (`int`):
            Number of parallel attention heads.
        dropout (`nn.Module`):
            Dropout probability.
    """
    def __init__(self, num_q_channels: int, num_kv_channels: int, num_heads: int, dropout: float):
        super().__init__()
        self.attention = nn.MultiheadAttention(
            embed_dim=num_q_channels,
            num_heads=num_heads,
            kdim=num_kv_channels,
            vdim=num_kv_channels,
            dropout=dropout,
            batch_first=True,
        )

    def forward(self, x_q, x_kv, pad_mask=None, attn_mask=None):
        """
        Forward function.
        Args:
            x_q (`Tensor`):
                Query embeddings.
            x_kv (`Tensor`):
                Key embeddings. Key equals value.
            pad_mask (`int`):
                Padding mask.
            attn_mask (`nn.Module`):
                Attention mask.
        """
        return self.attention(x_q, x_kv, x_kv, key_padding_mask=pad_mask, attn_mask=attn_mask)[0]



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import pathlib
from setuptools import setup

from typing import List


class DevelopCommand:
    """
    Implementation for subcmd `towhee develop`
    """

    def __init__(self, args) -> None:
        self._args = args
        print(args)

    def __call__(self):
        os.chdir(self._args.path)
        cwd = os.getcwd()
        has_pytorch = os.path.isdir('./pytorch')
        operator_name = os.path.basename(cwd).replace('-', '_')
        package_name = 'towheeoperator.{}_{}'.format(self._args.namespace,
                                                     operator_name)
        requirements = self.read_requirements()
        if self._args.pack:
            sys.argv = 'setup.py bdist_wheel --universal'.split()
        if self._args.install:
            sys.argv = ['setup.py', 'install']
        if self._args.develop:
            sys.argv = ['setup.py', 'develop']
            if not os.path.isdir('./towheeoperator'):
                os.mkdir('towheeoperator')
            link_target = 'towheeoperator/{}_{}'.format(
                self._args.namespace, operator_name)
            if not os.path.islink(link_target):
                os.symlink('..', link_target)
        setup(name=package_name,
              packages=[package_name, package_name +
                        '.pytorch'] if has_pytorch else [package_name],
              package_dir={package_name: '.'},
              package_data={'': ['*.txt', '*.rst']},
              install_requires=requirements,
              zip_safe=False)

    def read_requirements(self) -> List[str]:
        try:
            with open('requirements.txt', encoding='utf-8') as f:
                required = f.read().splitlines()
        except Exception:  # pylint: disable=broad-except)
            required = []
        return required

    @staticmethod
    def install(subparsers):
        parser_develop = subparsers.add_parser('develop',
                                               help='develop op/pipeline')
        group = parser_develop.add_mutually_exclusive_group(required=True)
        group.add_argument('-d',
                           '--develop',
                           action='store_true',
                           help='install `egg-link` file for development')
        group.add_argument('-p',
                           '--pack',
                           action='store_true',
                           help='pack the op/pipeline with `wheel` format')
        group.add_argument('-i',
                           '--install',
                           action='store_true',
                           help='install the op/pipeline')
        parser_develop.add_argument('-n',
                                    '--namespace',
                                    type=str,
                                    default='towhee',
                                    help='operator namespace')
        parser_develop.add_argument('path',
                                    type=pathlib.Path,
                                    nargs='?',
                                    default='.',
                                    help='root path of op/pipeline')



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List, Any

from towhee.dag.base_repr import BaseRepr


class OperatorRepr(BaseRepr):
    """This class encapsulates operator representations at compile-time.

    Args:
        name (`str`):
            Name of the operator represented by this object.
        function (`str`):
            The path leads to the operator.
        init_args (`Dict[str, any]`):
            The args to initilize the operator.
        inputs (`List[Dict[str, Any]]`):
            Input dataframes(s) to this object.
        outputs (`List[Dict[str, Any]]`):
            This operator's output dataframe.
        iter_info (`Dict[str, Any]`):
            This operator's iterator info.
    """

    def __init__(
        self,
        name: str,
        function: str,
        init_args: Dict[str, Any],
        inputs: List[Dict[str, Any]],
        outputs: List[Dict[str, Any]],
        iter_info: Dict[str, Any],
        tag: str = 'main',
        threads: int = 1

    ):
        super().__init__(name)
        self._function = function
        self._tag = tag
        self._inputs = inputs
        self._outputs = outputs
        self._init_args = init_args
        self._iter_info = iter_info
        self._threads = threads
        if self._threads != 1 and self._iter_info.get('type') not in ['filter', 'map']:
            raise RuntimeError('Only filter and map operator support multi-threads')

    @property
    def function(self):
        return self._function

    @property
    def inputs(self) -> List[dict]:
        """
        Returns:
            (`List[dict]`)
                The inputs of the operator.
        """
        return self._inputs

    @property
    def outputs(self) -> List:
        """
        Returns:
            (`List[dict]`)
                The outputs of the operator.
        """
        return self._outputs

    @property
    def init_args(self) -> Dict[str, Any]:
        """
        Returns:
            (`Dict[str, Any]`)
                The args to initilize the operator.
        """
        return self._init_args

    @property
    def tag(self) -> str:
        """
        Returns:
            (`str`)
                The tag to load of the operator.
        """
        return self._tag

    @property
    def iter_info(self) -> Dict[str, Any]:
        """
        Returns:
            (` Dict[str, Any]`)
                The operator's iterator info.
        """
        return self._iter_info

    @property
    def threads(self) -> int:
        return self._threads

    @staticmethod
    def from_dict(info: Dict[str, Any]) -> 'OperatorRepr':
        """
        Generate a OperatorRepr from a description dict.

        Args:
            info (`Dict[str, Any]`):
                A dict to describe the Operator.

        Returns:
            (`towhee.dag.OperatorRepr`)
                The OperatorRepe object.
        """
        if not BaseRepr.is_valid(info, {'name', 'init_args', 'function', 'inputs', 'outputs', 'iter_info'}):
            raise ValueError('Invalid operator info.')

        if 'tag' not in info:
            info['tag'] = 'main'

        return OperatorRepr(info['name'], info['function'], info['init_args'],
                            info['inputs'], info['outputs'], info['iter_info'], info['tag'], info.get('threads', 1))

    @staticmethod
    def from_ir(function: str, init_args: Dict[str, Any]) -> 'OperatorRepr':
        return OperatorRepr('', function, init_args, None, None, None)



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod

from towhee.dataframe.iterators import MapIterator
from towhee.types._frame import _Frame


class OutputFormat(ABC):
    """
    Format pipeline's output.
    """

    @abstractmethod
    def __call__(self, itr: MapIterator):
        pass

    @staticmethod
    def get_format_handler(pipeline_type: str):
        if pipeline_type == 'image-embedding':
            return ImageEmbeddingFormat()
        else:
            return NormalFormat()


class ImageEmbeddingFormat(OutputFormat):
    """
    Image embedding pipeline format.

    If pipelie's type is image-embedding, use the pipeline.
    """

    def __call__(self, itr: MapIterator):
        data = next(itr)
        return data[0][0]


class NormalFormat(OutputFormat):
    """
    Normal format.
    """

    def __call__(self, itr: MapIterator):
        res = []
        for data in itr:
            # data is Tuple
            data_value = []
            for item in data[0]:
                if item is not None and not isinstance(item, _Frame):
                    data_value.append(item)
            res.append(tuple(data_value))
        return res



# original code from https://github.com/SwinTransformer/Video-Swin-Transformer
# modified by Zilliz.

from torch import nn
import numpy as np
from einops import rearrange

from towhee.models.layers.swin_transformer_block3d import SwinTransformerBlock3D
from towhee.models.utils.get_window_size import get_window_size
from towhee.models.video_swin_transformer.compute_mask import compute_mask


class VideoSwinTransformerBlock(nn.Module):
    """
    A basic Swin Transformer Block for one stage.

    Args:
        dim (`int`):
            Number of feature channels
        depth (`int`):
            Depths of this stage.
        num_heads (`int`):
            Number of attention head.
        window_size (`tuple[int]`):
            Local window size. Default: (1,7,7).
        mlp_ratio (`float`):
            Ratio of mlp hidden dim to embedding dim. Default: 4.
        qkv_bias (`bool`):
            If True, add a learnable bias to query, key, value. Default: True
        qk_scale (`float`):
            Override default qk scale of head_dim ** -0.5 if set.
        drop (`float`):
            Dropout rate. Default: 0.0
        attn_drop (`float`):
            Attention dropout rate. Default: 0.0
        drop_path (`float`):
            Stochastic depth rate. Default: 0.0
        norm_layer (`nn.Module`):
            Normalization layer. Default: nn.LayerNorm
        downsample (`nn.Module`):
            Downsample layer at the end of the layer. Default: None
    """

    def __init__(self,
                 dim,
                 depth,
                 num_heads,
                 window_size=(1, 7, 7),
                 mlp_ratio=4.,
                 qkv_bias=False,
                 qk_scale=None,
                 drop=0.,
                 attn_drop=0.,
                 drop_path=0.,
                 norm_layer=nn.LayerNorm,
                 downsample=None,
                 use_checkpoint=False):
        super().__init__()
        self.window_size = window_size
        self.shift_size = tuple(i // 2 for i in window_size)
        self.depth = depth
        self.use_checkpoint = use_checkpoint

        # build blocks
        self.blocks = nn.ModuleList([
            SwinTransformerBlock3D(
                dim=dim,
                num_heads=num_heads,
                window_size=window_size,
                shift_size=(0, 0, 0) if (i % 2 == 0) else self.shift_size,
                mlp_ratio=mlp_ratio,
                qkv_bias=qkv_bias,
                qk_scale=qk_scale,
                drop=drop,
                attn_drop=attn_drop,
                drop_path=drop_path[i] if isinstance(drop_path, list) else drop_path,
                norm_layer=norm_layer,
                use_checkpoint=use_checkpoint,
            )
            for i in range(depth)])

        self.downsample = downsample
        if self.downsample is not None:
            self.downsample = downsample(dim=dim, norm_layer=norm_layer)

    def forward(self, x):
        """
        Forward function.

        Args:
            x (`torch.Tensor`):
                Input feature, tensor size (b, c, d, h, w).
        """
        # calculate attention mask for SW-MSA
        b, _, d, h, w = x.shape
        window_size, shift_size = \
            get_window_size((d, h, w), self.window_size, self.shift_size)  # pylint: disable=unbalanced-tuple-unpacking
        x = rearrange(x, 'b c d h w -> b d h w c')
        dp = int(np.ceil(d / window_size[0])) * window_size[0]
        hp = int(np.ceil(h / window_size[1])) * window_size[1]
        wp = int(np.ceil(w / window_size[2])) * window_size[2]
        attn_mask = compute_mask(dp, hp, wp, window_size, shift_size, x.device)
        for blk in self.blocks:
            x = blk(x, attn_mask)
        x = x.view(b, d, h, w, -1)

        if self.downsample is not None:
            x = self.downsample(x)
        x = rearrange(x, 'b d h w c -> b c d h w')
        return x



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import subprocess
import sys
import time
from typing import Tuple, List, Union
from shutil import rmtree
from pathlib import Path
from requests.exceptions import HTTPError

from tqdm import tqdm
from tempfile import TemporaryFile
from concurrent.futures import ThreadPoolExecutor
from towhee.utils.hub_utils import HubUtils
from towhee.utils.log import engine_log


class RepoManager:
    """
    Base Repo Manager class to create, initialize, download repos.

    Args:
        author (`str`):
            The author of the repo.
        repo (`str`):
            The name of the repo.
        root (`str`):
            The root url of the repo.
    """
    def __init__(self, author: str, repo: str, root: str = 'https://towhee.io'):
        self._author = author
        self._repo = repo
        self._root = root
        self._temp = {'pipeline': 'pipeline-template', 'pyoperator': 'pyoperator-template', 'nnoperator': 'nnoperator-template'}
        self.hub_utils = HubUtils(self._author, self._repo, self._root)

    @property
    def author(self):
        return self._author

    @property
    def repo(self):
        return self._repo

    @property
    def root(self):
        return self._root

    def exists(self) -> bool:
        """
        Check if a repo exists.

        Returns:
            (`bool`)
                return `True` if the repository exists, else `False`.
        """
        res = self.hub_utils.get_info()
        return res.status_code == 200

    def obtain_lfs_extensions(self, tag: str) -> List[str]:
        """
        Download the .gitattributes file from the specified repo in order to figure out
        which files are being tracked by git-lfs.

        Lines that deal with git-lfs take on the following format:

        ```
            *.extension   filter=lfs  merge=lfs ...
        ```

        Args:
            tag (`str`):
                The tag name.

        Returns:
            (`List[str]`)
                The list of file extentions tracked by git-lfs.
        """
        res = self.hub_utils.get_file('.gitattributes', tag)
        lfs_files = []
        if res.status_code == 200:
            # Using temporary file in order to avoid double download, cleaner to not split up downloads everywhere.
            with TemporaryFile() as temp_file:
                temp_file.write(res.content)
                temp_file.seek(0)

                for line in temp_file:
                    parts = line.split()
                    # We only care if lfs filter is present.
                    if b'filter=lfs' in parts[1:]:
                        # Removing the `*` in `*.ext`, need work if filtering specific files.
                        lfs_files.append(parts[0].decode('utf-8')[1:])
        return lfs_files

    def latest_commit(self, tag: str) -> str:
        """
        Grab the latest commit of a tag.

        Args:
            tag (`str`):
                The tag name.

        Returns:
            (`str`)
                The latest commit hash cut down to 10 characters.

        Raises:
            (`HTTPError`)
                Raise error in request.
        """
        res = self.hub_utils.get_commits(limit=1, page=1, tag=tag)
        return res.json()[0]['sha'][:10]

    def get_file_list(self, commit: str) -> List[str]:
        """
        Get all the files in the current repo at the given commit.

        This is done through forming a git tree recursively and filtering out all the files.

        Args:
            commit (`str`):
                The commit to base current existing files.

        Returns:
            (`List[str]`)
                The file paths for the repo

        Raises:
            (`HTTPError`)
                Raise error in request.
        """
        file_list = []
        res = self.hub_utils.get_tree(commit)

        # Check each object in the tree
        for file in res.json()['tree']:
            # Ignore directories (they have the type 'tree')
            if file['type'] != 'tree':
                file_list.append(file['path'])
        return file_list

    def download_executor(self, tag: str, file_name: str, lfs_files: Tuple[str], local_repo_path: Union[str, Path]) -> None:
        """
        Load the content from url and write into local files.

        Args:
            tag (`str`):
                The tag name.
            file_name (`str`):
                The hub file paths.
            lfs_files (`str):
                The file extensions being tracked by git-lfs.
            local_repo_path (`Union[str, Path]`):
                The local directory to download to.

        Raises:
            (`HTTPError`)
                Rasie error in request.
            (`OSError`)
                Raise error in writing file.
        """
        file_path = Path(local_repo_path) / file_name
        if not file_path.parent.resolve().exists():
            try:
                file_path.parent.resolve().mkdir(parents=True)
            except FileExistsError:
                pass
            except OSError as e:
                raise e

        # Get content.
        for i in range(5):
            try:
                if file_name.endswith(lfs_files):
                    res = self.hub_utils.get_lfs(file_name, tag)
                else:
                    res = self.hub_utils.get_file(file_name, tag)

                if res.status_code == 200:
                    break
                elif res.status_code == 429 and i < 4:
                    time.sleep(2)
                    continue
                res.raise_for_status()
            except HTTPError as e:
                raise e

        # Create local files.
        file_size = int(res.headers.get('content-length', 0))
        chunk_size = 1024
        progress_bar = tqdm(total=file_size, unit='iB', unit_scale=True, desc=f'Downloading {file_name}')
        with open(file_path, 'wb') as local_file:
            for chunk in res.iter_content(chunk_size=chunk_size):
                local_file.write(chunk)
                progress_bar.update(len(chunk))
        progress_bar.close()

    def download_files(self, tag: str, file_list: List[str], lfs_files: List[str], local_repo_path: Union[str, Path], install_reqs: bool) -> None:
        """
        Download the given files.

        Agrs:
            tag (`str`):
                The tag of the repo to download.
            file_list (List[str]):
                The files to download
            lfs_files (List[str]):
                The lfs files extensions.
            local_repo_path (`Union[str, Path]`):
                Thre local dir to download the files into.
            install_reqs (`bool`):
                Whether to install packages from requirements.txt.
        """
        local_repo_path = Path(local_repo_path)

        # endswith() can check multiple suffixes if they are a tuple.
        lfs_files = tuple(lfs_files)
        futures = []

        temp_path = local_repo_path.parent / 'dl-temp'
        with ThreadPoolExecutor(max_workers=50) as pool:
            for file_name in file_list:
                futures.append(pool.submit(self.download_executor, tag, file_name, lfs_files, temp_path))

        try:
            _ = [i.result() for i in futures]
            temp_path.rename(local_repo_path)
        except Exception as e:
            rmtree(temp_path)
            raise e

        if install_reqs:
            requirements = list(filter(lambda x: re.match(r'(.*/)?requirements.txt', x) is not None, file_list))
            for req in requirements:
                subprocess.check_call([sys.executable, '-m', 'pip', 'install', '-r', str(local_repo_path / req)])

    def download(self, local_repo_path: Union[str, Path] = Path.cwd(), tag: str = 'main', install_reqs: bool = True) -> None:
        """
        Download repo without git.

        Agrs:
            local_repo_path (`Union[str, Path]`):
                Thre local dir to download the files into.
            tag (`str`):
                The tag of the repo to download.
            install_reqs (`bool`):
                Whether to install packages from requirements.txt.
        """
        if not self.exists():
            engine_log.error('%s/%s repo does not exist.', self._author, self._repo)
            raise ValueError(f'{self._author}/{self._author} repo does not exist.')

        local_repo_path = Path(local_repo_path)
        lfs_files = self.obtain_lfs_extensions(tag)
        commit = self.latest_commit(tag)
        file_list = self.get_file_list(commit)
        self.download_files(tag=tag, file_list=file_list, lfs_files=lfs_files, local_repo_path=local_repo_path, install_reqs=install_reqs)



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict

from towhee.engine.operator_runner import runner_base
from towhee.engine.operator_runner import map_runner
from towhee.engine.operator_runner import concat_runner
from towhee.engine.operator_runner import flatmap_runner
from towhee.engine.operator_runner import filter_runner
from towhee.engine.operator_runner import window_runner
from towhee.engine.operator_runner import generator_runner
from towhee.engine.operator_io.reader import DataFrameReader
from towhee.engine.operator_io.writer import DataFrameWriter


def create_runner(
    runner_type: str,
    name: str,
    index: int,
    op_name: str,
    tag: str,
    hub_op_id: str,
    op_args: Dict[str, any],
    reader: DataFrameReader,
    writer: DataFrameWriter,
) -> runner_base.RunnerBase:
    if runner_type.lower() == 'map':
        return map_runner.MapRunner(name, index, op_name, tag, hub_op_id, op_args, reader, writer)
    elif runner_type.lower() == 'flatmap':
        return flatmap_runner.FlatMapRunner(name, index, op_name, tag, hub_op_id, op_args, reader, writer)
    elif runner_type.lower() == 'filter':
        return filter_runner.FilterRunner(name, index, op_name, tag, hub_op_id, op_args, reader, writer)
    elif runner_type.lower() == 'concat':
        return concat_runner.ConcatRunner(name, index, op_name, tag, hub_op_id, op_args, reader, writer)
    elif runner_type.lower() in ['window', 'time_window']:
        return window_runner.WindowRunner(name, index, op_name, tag, hub_op_id, op_args, reader, writer)
    elif runner_type.lower() == 'generator':
        return generator_runner.GeneratorRunner(name, index, op_name, tag, hub_op_id, op_args, reader, writer)
    else:
        raise AttributeError('No runner type named: {}'.format(runner_type))



# Copyright 2021 Ross Wightman . All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# This code is modified by Zilliz.

from torch import nn

def hard_mish(x, inplace: bool = False):
    if inplace:
        return x.mul_(0.5 * (x + 2).clamp(min=0, max=2))
    else:
        return 0.5 * x * (x + 2).clamp(min=0, max=2)

class HardMish(nn.Module):
    """
    Hard Mish
    Experimental, based on notes by Mish author Diganta Misra at
      https://github.com/digantamisra98/H-Mish/blob/0da20d4bc58e696b6803f2523c58d3c8a82782d0/README.md
    Args:
        inplace(`Bool`):
            whether use inplace version.
    Returns:
        (`torch.Tensor`)
            output tensor after activation.
    """
    def __init__(self, inplace: bool = False):
        super().__init__()
        self.inplace = inplace

    def forward(self, x):
        return hard_mish(x, self.inplace)



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class ComputerVisionMixin:
    """
    Mixin for computer vision problems.
    """

    # pylint: disable=import-outside-toplevel
    def image_imshow(self, title='image'):
        from towhee.utils.cv2_utils import cv2
        for im in self:
            cv2.imshow(title, im)
            cv2.waitKey(1)


if __name__ == '__main__':
    import doctest
    doctest.testmod(verbose=False)



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from towhee.engine.factory import ops


class DispatcherMixin:
    """
    Mixin for call dispatcher for data collection

    >>> from towhee import register
    >>> from towhee import ops
    >>> from towhee import DataCollection
    >>> @register(name='add_1')
    ... def add_1(x):
    ...     return x+1

    >>> dc = DataCollection.range(5).stream()
    >>> dc.add_1['a','b','c']()
    """

    def resolve(self, call_mapping, path, index, *arg, **kws):
        # pylint: disable=protected-access
        if path in call_mapping:
            return call_mapping[path](*arg, **kws)
        else:
            return ops._callback(path, index, *arg, **kws)


if __name__ == '__main__':
    import doctest

    doctest.testmod(verbose=False)



# Copyright 2021  Facebook. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# This code is modified by Zilliz.


import torch
from torch import nn


class SequencePool(nn.Module):
    """
    Sequence pool produces a single embedding from a sequence of embeddings.
    Currently it supports "mean" and "cls".
    """

    def __init__(self, mode: str) -> None:
        """
        Args:
            mode ('str'):
                If set to "cls", it assumes the first element in the input is the cls token and returns it.
                If set to "mean", it returns the mean of the entire sequence.
        """
        super().__init__()
        self.mode = mode
        assert mode in ["cls", "mean"], "Unsupported mode for SequencePool."

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if self.mode == "cls":
            x = x[:, 0]
        elif self.mode == "mean":
            x = x.mean(1)
        else:
            raise NotImplementedError
        return x



# Copyright 2021  Facebook. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# This code is modified by Zilliz.


import torch
from torch import nn
from typing import List, Tuple
import numpy

from towhee.models.layers.pool_attention import AttentionPool


class MultiScaleAttention(nn.Module):
    """
    A multiscale attention block.
    compare to a conventional attention block, a multiscale attention block optionally
    supports pooling (either before or after qkv projection). If pooling is not used, a
    multiscale attention block is equivalent to a conventional attention block.
    ::
                                   Input
                                     |
                    |----------------|-----------------|
                    ↓                ↓                 ↓
                  Linear           Linear            Linear
                    &                &                 &
                 Pool (Q)         Pool (K)          Pool (V)
                    → -------------- ←                 |
                             ↓                         |
                       MatMul & Scale                  |
                             ↓                         |
                          Softmax                      |
                             → ----------------------- ←
                                         ↓
                                   MatMul & Scale
                                         ↓
                                      DropOut
    Args:
        dim(int):
            Input feature dimension.
        num_heads(int):
            number of heads in the attention layer.
        qkv_bias(bool):
            If set to False, the qkv layer will not learn an additive bias.
        dropout_rate(float):
            Dropout rate.
        kernel_q(_size_3_t):
            Pooling kernel size for q. If both pooling kernel
            size and pooling stride size are 1 for all the dimensions, pooling is disabled.
        kernel_kv(_size_3_t):
            Pooling kernel size for kv. If both pooling kernel size and pooling stride size
            are 1 for all the dimensions, pooling is disabled.
        stride_q(_size_3_t):
            Pooling kernel stride for q.
        stride_kv(_size_3_t):
            Pooling kernel stride for kv.
        norm_layer(nn.Module):
            normalization layer used after pooling.
        has_cls_embed(bool):
            If set to True, the first token of the input tensor
            should be a cls token. Otherwise, the input tensor does not contain a cls token.
            Pooling is not applied to the cls token.
        pool_mode(str):
            Pooling mode. Option includes "conv" (learned pooling), "avg"
            (average pooling), and "max" (max pooling).
        pool_first(bool):
            If set to True, pool is applied before qkv projection.
            Otherwise, pool is applied after qkv projection.
    """

    def __init__(
        self,
        dim,
        num_heads=8,
        qkv_bias=False,
        dropout_rate=0.0,
        kernel_q=(1, 1, 1),
        kernel_kv=(1, 1, 1),
        stride_q=(1, 1, 1),
        stride_kv=(1, 1, 1),
        norm_layer=nn.LayerNorm,
        has_cls_embed=True,
        pool_mode=nn.Conv3d,
        pool_first=False,
    ) -> None:
        super().__init__()
        self.dim = dim
        self.num_heads = num_heads
        self.qkv_bias = qkv_bias
        self.dropout_rate = dropout_rate
        self.kernel_q = kernel_q
        self.kernel_kv = kernel_kv
        self.stride_q = stride_q
        self.stride_kv = stride_kv
        self.norm_layer = norm_layer
        self.has_cls_embed = has_cls_embed
        self.pool_mode = pool_mode
        self.pool_first = pool_first

        assert self.pool_mode in [nn.Conv3d, nn.AvgPool3d, nn.MaxPool3d]

        self.head_dim = self.dim // self.num_heads
        self.scale = self.head_dim ** -0.5
        self.padding_q = [int(q // 2) for q in self.kernel_q]
        self.padding_kv = [int(kv // 2) for kv in self.kernel_kv]

        self.q = nn.Linear(self.dim, self.dim, bias=self.qkv_bias)
        self.k = nn.Linear(self.dim, self.dim, bias=self.qkv_bias)
        self.v = nn.Linear(self.dim, self.dim, bias=self.qkv_bias)
        self.proj = nn.Linear(self.dim, self.dim)
        if self.dropout_rate > 0.0:
            self.proj_drop = nn.Dropout(self.dropout_rate)

        # Skip pooling with kernel and stride size of (1, 1, 1).
        if (
            self.kernel_q is not None
            and numpy.prod(self.kernel_q) == 1
            and numpy.prod(self.stride_q) == 1
        ):
            self.kernel_q = None
        if (
            self.kernel_kv is not None
            and numpy.prod(self.kernel_kv) == 1
            and numpy.prod(self.stride_kv) == 1
        ):
            self.kernel_kv = None

        if self.pool_mode in (nn.AvgPool3d, nn.MaxPool3d):
            pool_op = nn.MaxPool3d if pool_mode == nn.MaxPool3d else nn.AvgPool3d
            self.pool_q = (
                pool_op(self.kernel_q, self.stride_q, self.padding_q, ceil_mode=False)
                if self.kernel_q is not None
                else None
            )
            self.pool_k = (
                pool_op(self.kernel_kv, self.stride_kv, self.padding_kv, ceil_mode=False)
                if self.kernel_kv is not None
                else None
            )
            self.pool_v = (
                pool_op(self.kernel_kv, self.stride_kv, self.padding_kv, ceil_mode=False)
                if self.kernel_kv is not None
                else None
            )
        elif self.pool_mode == nn.Conv3d:
            self.pool_q = (
                nn.Conv3d(
                    self.head_dim,
                    self.head_dim,
                    self.kernel_q,
                    stride=self.stride_q,
                    padding=self.padding_q,
                    groups=self.head_dim,
                    bias=False,
                )
                if self.kernel_q is not None
                else None
            )
            self.norm_q = self.norm_layer(self.head_dim) if self.kernel_q is not None else None
            self.pool_k = (
                nn.Conv3d(
                    self.head_dim,
                    self.head_dim,
                    self.kernel_kv,
                    stride=self.stride_kv,
                    padding=self.padding_kv,
                    groups=self.head_dim,
                    bias=False,
                )
                if self.kernel_kv is not None
                else None
            )
            self.norm_k = self.norm_layer(self.head_dim) if self.kernel_kv is not None else None
            self.pool_v = (
                nn.Conv3d(
                    self.head_dim,
                    self.head_dim,
                    self.kernel_kv,
                    stride=self.stride_kv,
                    padding=self.padding_kv,
                    groups=self.head_dim,
                    bias=False,
                )
                if self.kernel_kv is not None
                else None
            )
            self.norm_v = self.norm_layer(self.head_dim) if self.kernel_kv is not None else None
        else:
            raise NotImplementedError("Unsupported model.")

    def qkv_proj(
            self,
            q,
            q_size,
            k,
            k_size,
            v,
            v_size,
            batch_size,
            chan_size,
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        Args:
            q(torch.Tensor):
                q tensor.
            q_size(List[int]):
                q tensor size.
            k(torch.Tensor):
                k tensor.
            k_size(List[int]):
                k tensor size.
            v(torch.Tensor):
                v tensor.
            v_size(List[int]):
                v tensor size.
            batch_size(List[int]):
                batch size.
            chan_size(List[int]):
                channel size.
        """
        q = (
            self.q(q)
                .reshape(batch_size, q_size, self.num_heads, chan_size // self.num_heads)
                .permute(0, 2, 1, 3)
        )
        k = (
            self.k(k)
                .reshape(batch_size, k_size, self.num_heads, chan_size // self.num_heads)
                .permute(0, 2, 1, 3)
        )
        v = (
            self.v(v)
                .reshape(batch_size, v_size, self.num_heads, chan_size // self.num_heads)
                .permute(0, 2, 1, 3)
        )
        return q, k, v

    def qkv_pool(
            self,
            q: torch.Tensor,
            k: torch.Tensor,
            v: torch.Tensor,
            thw_shape: Tuple[torch.Tensor, List[int]],
    ) -> Tuple[
        torch.Tensor, List[int], torch.Tensor, List[int], torch.Tensor, List[int]
    ]:
        """
        Args:
            q(torch.Tensor):
                q tensor.
            k(torch.Tensor):
                k tensor.
            v(torch.Tensor):
                v tensor.
            thw_shape(Tuple[torch.Tensor, List[int]]):
                The shape of the input tensor.
        """
        ap = AttentionPool(
            pool=self.pool_q,
            thw_shape=thw_shape,
            has_cls_embed=self.has_cls_embed,
            norm=self.norm_q if hasattr(self, "norm_q") else None,
        )
        q, q_shape = ap(q)
        ap = AttentionPool(
            pool=self.pool_k,
            thw_shape=thw_shape,
            has_cls_embed=self.has_cls_embed,
            norm=self.norm_k if hasattr(self, "norm_k") else None,
        )
        k, k_shape = ap(k)
        ap = AttentionPool(
            pool=self.pool_v,
            thw_shape=thw_shape,
            has_cls_embed=self.has_cls_embed,
            norm=self.norm_v if hasattr(self, "norm_v") else None,
        )
        v, v_shape = ap(v)
        return q, q_shape, k, k_shape, v, v_shape

    def get_qkv_length(
            self,
            q_shape,
            k_shape,
            v_shape,
    ) -> Tuple[int]:
        """
        Args:
            q_shape(List[int]):
                q tensor shape.
            k_shape(List[int]):
                k tensor shape.
            v_shape(List[int]):
                v tensor shape.
        """
        q_n = numpy.prod(q_shape) + 1 if self.has_cls_embed else numpy.prod(q_shape)
        k_n = numpy.prod(k_shape) + 1 if self.has_cls_embed else numpy.prod(k_shape)
        v_n = numpy.prod(v_shape) + 1 if self.has_cls_embed else numpy.prod(v_shape)
        return q_n, k_n, v_n

    def reshape_qkv_to_seq(
            self,
            q,
            k,
            v,
            q_n,
            v_n,
            k_n,
            b,
            c,
    ) -> Tuple[int]:
        """
        Args:
            q(torch.Tensor):
                q tensor.
            k(torch.Tensor):
                k tensor.
            v(torch.Tensor):
                v tensor.
            q_n(int):
                k tensor size.
            v_n(int):
                v tensor size.
            k_n(int):
                k tensor size.
            b(int):
                Reshaped size.
            c(int):
                Reshaped size.
        """
        q = q.permute(0, 2, 1, 3).reshape(b, q_n, c)
        v = v.permute(0, 2, 1, 3).reshape(b, v_n, c)
        k = k.permute(0, 2, 1, 3).reshape(b, k_n, c)
        return q, k, v

    def forward(
        self, x: torch.Tensor, thw_shape: List[int]
    ) -> Tuple[torch.Tensor, List[int]]:
        """
        Args:
            x(torch.Tensor):
                Input tensor.
            thw_shape(List):
                The shape of the input tensor (before flattening).
        """

        b, n, c = x.shape
        if self.pool_first:
            x = x.reshape(b, n, self.num_heads, c // self.num_heads).permute(0, 2, 1, 3)
            q = k = v = x
            q, q_shape, k, k_shape, v, v_shape = self.qkv_pool(q, k, v, thw_shape)
            q_n, k_n, v_n = self.get_qkv_length(q_shape, k_shape, v_shape)
            q, k, v = self.reshape_qkv_to_seq(q, k, v, q_n, v_n, k_n, b, c)
            q, k, v = self.qkv_proj(q, q_n, k, k_n, v, v_n, b, c)
        else:
            q = k = v = x
            q, k, v = self.qkv_proj(q, n, k, n, v, n, b, c)
            q, q_shape, k, k_shape, v, v_shape = self.qkv_pool(q, k, v, thw_shape)

        attn = (q @ k.transpose(-2, -1)) * self.scale
        attn = attn.softmax(dim=-1)

        n = q.shape[2]
        x = (attn @ v).transpose(1, 2).reshape(b, n, c)
        x = self.proj(x)
        if self.dropout_rate > 0.0:
            x = self.proj_drop(x)
        return x, q_shape



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from typing import List
import random
import weakref
import threading

from towhee.engine.thread_pool_task_executor import ThreadPoolTaskExecutor


class TaskScheduler(ABC):
    """Task scheduler abstract interface.

    Args:
        task_execs: (`List[towhee.TaskExecutor]`)
            A list of task executors that the `Engine` manages. Also should be
            continuously changing as new executors are acquired.
    """

    def __init__(self, task_execs: List[ThreadPoolTaskExecutor]):
        self._task_execs = task_execs
        self._graph_ctx_refs = []
        self._need_stop = False
        self._event = threading.Event()

    @abstractmethod
    def stop(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def join(self) -> None:
        raise NotImplementedError

    def schedule_forever(self, sleep_ms):
        """Runs the a single schedule step in a loop.

        sleep_ms: (`int`)
            Milliseconds to sleep after completing a single scheduling step.
        """
        while not self._need_stop:
            self.schedule_step()
            self._event.wait(sleep_ms / 1000)

    @abstractmethod
    def schedule_step(self):
        raise NotImplementedError


class BasicScheduler(TaskScheduler):
    """
    Basic scheduler.

    This scheduler will start all ops, and has no scheduling logic, so the
    schedule_step does nothing.
    """

    def __init__(self, task_execs: List[ThreadPoolTaskExecutor], threshold: int):
        super().__init__(task_execs)
        self._lock = threading.Lock()
        self._df_threshold = threshold

    def register(self, graph_ctx):
        for op in graph_ctx.op_ctxs.values():
            executor = self._find_optimal_exec()
            op.start(executor)

        with self._lock:
            self._graph_ctx_refs.append(weakref.ref(graph_ctx))

    def stop(self):
        with self._lock:
            for g_ctx_ref in self._graph_ctx_refs:
                g_ctx = g_ctx_ref()
                if g_ctx is not None:
                    g_ctx.stop()
        self._event.set()

    def join(self):
        with self._lock:
            for g_ctx_ref in self._graph_ctx_refs:
                g_ctx = g_ctx_ref()
                if g_ctx is not None:
                    g_ctx.join()

    def _remove_finished_graph(self):
        start_index = 0
        with self._lock:
            for g_ctx_ref in self._graph_ctx_refs:
                if g_ctx_ref() is not None:
                    break
                start_index += 1
            if start_index != 0:
                self._graph_ctx_refs = self._graph_ctx_refs[start_index:]

    def _scheduler_ops(self):
        # threshold <=0: disable scheduler
        if self._df_threshold <= 0:
            return

        for g_ctx_ref in self._graph_ctx_refs:
            g_ctx = g_ctx_ref()
            if g_ctx is not None:
                for name, df in g_ctx.dataframes.items():
                    if df.current_size > self._df_threshold:
                        g_ctx.slow_down(name, df.current_size / self._df_threshold)
                    elif df.current_size == 0 and not df.sealed:
                        g_ctx.speed_up(name)
                    else:
                        pass

    def _df_gc(self):
        for g_ctx_ref in self._graph_ctx_refs:
            g_ctx = g_ctx_ref()
            if g_ctx is not None:
                g_ctx.gc()

    def schedule_step(self):
        """
        Scheduler:
            1. Remove the graph which is finished.
            2. Call df gc.

        """
        self._remove_finished_graph()
        self._df_gc()
        self._scheduler_ops()

    def _find_optimal_exec(self):
        """
        Acquires the least busy instance of `TaskExecutor` that can still execute the operator.
        """
        return random.choice(self._task_execs)



# Copyright 2021 Ross Wightman . All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# This code is modified by Zilliz.
from typing import Tuple, Optional, Union

import torch
from torch import nn
import torch.nn.functional as F
from torch.nn.common_types import _size_2_t

from towhee.models.layers.padding_functions import pad_same

def conv2d_same(
       x: torch.Tensor, weight: torch.Tensor, bias: Optional[torch.Tensor] = None, stride: Tuple[int, int] = (1, 1),
        padding: Tuple[int, int] = (0, 0), dilation: Tuple[int, int] = (1, 1), groups: int = 1):
    """
    Tensorflow like 'SAME' convolution function for 2D convolutions.
    """
    x = pad_same(x, weight.shape[-2:], stride, dilation)
    _ = padding
    return F.conv2d(x, weight, bias, stride, (0, 0), dilation, groups)

class Conv2dSame(nn.Conv2d):
    """
    Tensorflow like 'SAME' convolution wrapper for 2D convolutions.

    Args:
        in_channels (`int`):
            Number of channels in the input image.
        out_channels (`int`):
            Number of channels produced by the convolution.
        kernel_size (`Union[int, Tuple]`):
            Size of the convolving kernel.
        stride (`Union[int, Tuple]`):
            Stride of the convolution.
        padding (`Union[int, Tuple, str]`):
            Padding added to all four sides of the input.
        dilation (`int`):
            Spacing between kernel elements.
        groups (`int`):
            Number of blocked connections from input channels to output channels.
        bias (`bool`):
            If True, adds a learnable bias to the output.
    """

    def __init__(self,
        in_channels: int,
        out_channels: int,
        kernel_size: _size_2_t,
        stride: _size_2_t = 1,
        padding: Union[str, _size_2_t] = 0,
        dilation: _size_2_t = 1,
        groups: int = 1,
        bias: bool = True) -> None:
        super().__init__(
            in_channels, out_channels, kernel_size, stride, 0, dilation, groups, bias)
        _ = padding

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return conv2d_same(x, self.weight, self.bias, self.stride, self.padding, self.dilation, self.groups)




# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy
from pathlib import Path
from typing import Union, Any

from towhee import pipeline


class ExecuteCommand:
    """
    Implementation for subcmd `towhee run`
    """
    def __init__(self, args) -> None:
        self._args = args

    def __call__(self):
        pipe = pipeline(self._args.pipeline)
        res = pipe(self._args.input)
        if not self._args.output:
            print(res)
        else:
            path = Path(self._args.output)
            self.save_result(path, res)

    @staticmethod
    def install(subparsers):
        parser_execute = subparsers.add_parser('run', help='execute command: run towhee pipeline')

        parser_execute.add_argument('-i', '--input', required=True, help='input the parameter for pipeline defaults to None')
        parser_execute.add_argument('-o', '--output', default=None, help='optional, path to the file that will be used to write results], '
                                                                         'defaults to None which will print the result')
        parser_execute.add_argument('pipeline', type=str, help='pipeline repo or path to yaml')

    @staticmethod
    def save_result(output: Union[str, Path], res: Any) -> None:
        """
        Save the results to local `output` file.

        Args:
            output (`str` | `Path`):
                The path that you are trying to save.
            res (`Any`):
                The result with any format.
        """
        file_name = Path(output) / 'towhee_output.txt'
        print(f'writing result to Path({str(output)})/towhee_output.txt')
        with open(str(file_name), 'w', encoding='utf-8') as f:
            if isinstance(res, list):
                f.write('[')
                for item in res:
                    f.write(f'{item}')
                f.write(']\n')
            elif isinstance(res, numpy.ndarray):
                numpy.savetxt(str(file_name), res)
            else:
                f.write(f'{str(res)}\n')



# Copyright 2021  Facebook. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# This code is modified by Zilliz.


from torch import nn

from towhee.models.layers.spatial_temporal_cls_positional_encoding import SpatialTemporalClsPositionalEncoding


def init_vit_weights(model, trunc_normal_std=0.02) -> None:
    """
    Weight initialization for vision transformers.
    Args:
        model(nn.Module):
            Model to be initialized.
        trunc_normal_std(float):
            The expected standard deviation for fully-connected layer and ClsPositionalEncoding.
    """
    for m in model.modules():
        if isinstance(m, nn.Linear):
            nn.init.trunc_normal_(m.weight, std=trunc_normal_std)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)
        elif isinstance(m, SpatialTemporalClsPositionalEncoding):
            for weights in m.parameters():
                nn.init.trunc_normal_(weights, std=trunc_normal_std)





# Copyright 2021  Facebook. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# This code is modified by Zilliz.


import torch
from torch import nn
from functools import partial

from towhee.models.layers.vision_transformer_basic_head import VisionTransformerBasicHead
from towhee.models.layers.patch_embed2d import PatchEmbed2D
from towhee.models.layers.spatial_temporal_cls_positional_encoding import SpatialTemporalClsPositionalEncoding
from towhee.models.utils.round_width import round_width
from towhee.models.layers.multi_scale_transformer_block import MultiScaleBlock
from towhee.models.multiscale_vision_transformers.multiscale_vision_transformers import MultiscaleVisionTransformers

def create_multiscale_vision_transformers(
    spatial_size,
    temporal_size,
    cls_embed_on=True,
    sep_pos_embed=True,
    depth=16,
    norm=nn.LayerNorm,
    enable_patch_embed=True,
    input_channels=3,
    patch_embed_dim=96,
    conv_patch_embed_kernel=(3, 7, 7),
    conv_patch_embed_stride=(2, 4, 4),
    conv_patch_embed_padding=(1, 3, 3),
    enable_patch_embed_norm=False,
    use_2d_patch=False,
    num_heads=1,
    mlp_ratio=4.0,
    qkv_bias=True,
    dropout_rate_block=0.0,
    droppath_rate_block=0.0,
    pooling_mode=nn.Conv3d,
    pool_first=False,
    embed_dim_mul=None,
    atten_head_mul=None,
    pool_q_stride_size=None,
    pool_kv_stride_size=None,
    pool_kv_stride_adaptive=None,
    pool_kvq_kernel=None,
    head=VisionTransformerBasicHead,
    head_dropout_rate=0.5,
    head_activation=None,
    head_num_classes=400
) -> nn.Module:
    """
    Multiscale Vision Transformers
    https://arxiv.org/abs/2104.11227
    ::
                                           PatchEmbed
                                               ↓
                                       PositionalEncoding
                                               ↓
                                            Dropout
                                               ↓
                                         Normalization
                                               ↓
                                             Block 1
                                               ↓
                                               .
                                               .
                                               .
                                               ↓
                                             Block N
                                               ↓
                                         Normalization
                                               ↓
                                              Head
    Args:
        spatial_size ('_size_2_t'):
            Input video spatial resolution (H, W). If a single int is given,
            it assumes the width and the height are the same.
        temporal_size ('int'):
            Number of frames in the input video.
        cls_embed_on ('bool'):
            If True, use cls embed in the model. Otherwise features are average
            pooled before going to the final classifier.
        sep_pos_embed ('bool'):
            If True, perform separate spatiotemporal embedding.
        depth ('int'):
            The depth of the model.
        norm ('Callable'):
            Normalization layer.
        enable_patch_embed ('bool'):
            If true, patchify the input video. If false, it assumes the input should
            have the feature dimension of patch_embed_dim.
        input_channels ('int'):
            Channel dimension of the input video.
        patch_embed_dim ('int'):
            Embedding dimension after patchifing the video input.
        conv_patch_embed_kernel ('Tuple[int]'):
            Kernel size of the convolution for patchifing the video input.
        conv_patch_embed_stride ('Tuple[int]'):
            Stride size of the convolution for patchifing the video input.
        conv_patch_embed_padding ('Tuple[int]'):
            Padding size of the convolution for patchifing the video input.
        enable_patch_embed_norm ('bool'):
            If True, apply normalization after patchifing the video input.
        use_2d_patch ('bool'):
            If True, use 2D convolutions to get patch embed. Otherwise, use 3D convolutions.
        num_heads ('int'):
            Number of heads in the first transformer block.
        mlp_ratio ('float'):
            MLP ratio which controls the feature dimension in the hidden layer of the Mlp block.
        qkv_bias ('bool'):
            If set to False, the qkv layer will not learn an additive bias.
        dropout_rate_block ('float'):
            Dropout rate for the attention block.
        droppath_rate_block ('float'):
            Droppath rate for the attention block.
        pooling_mode ('Callable'):
            Pooling mode.
        pool_first ('bool'):
            If set to True, pool is applied before qkv projection. Otherwise, pool is applied after qkv projection.
        embed_dim_mul ('List[List[int]]'):
            Dimension multiplication at layer i. If X is used, then the next block will increase
            the embed dimension by X times. Format: [depth_i, mul_dim_ratio].
        atten_head_mul ('List[List[int]]'):
            Head dimension multiplication at  layer i. If X is used, then the next block will increase
            the head by X times. Format: [depth_i, mul_dim_ratio].
        pool_q_stride_size ('List[List[int]]'):
            List of stride sizes for the pool q at each layer. Format: [[i, stride_t_i, stride_h_i, stride_w_i], ...,].
        pool_kv_stride_size ('List[List[int]]'):
            List of stride sizes for the pool kv at each layer. Format: [[i, stride_t_i, stride_h_i, stride_w_i], ...,].
        pool_kv_stride_adaptive ('_size_3_t'):
            Initial kv stride size for the first block. The stride size will be further reduced at the layer
            where q is pooled with the ratio of the stride of q pooling. If pool_kv_stride_adaptive is set,
            then pool_kv_stride_size should be none.
        pool_kvq_kernel ('_size_3_t'):
            Pooling kernel size for q and kv. It None, the kernel_size is [s + 1 if s > 1 else s for s in stride_size].
        head ('Callable'):
            Head model.
        head_dropout_rate ('float'):
            Dropout rate in the head.
        head_activation ('float'):
            Activation in the head.
        head_num_classes ('int'):
            Number of classes in the final classification head.
    """

    if use_2d_patch:
        assert temporal_size == 1, "If use_2d_patch, temporal_size needs to be 1."
    if pool_kv_stride_adaptive is not None:
        assert (
                pool_kv_stride_size is None
        ), "pool_kv_stride_size should be none if pool_kv_stride_adaptive is set."
    if norm == nn.LayerNorm:
        norm_layer = partial(nn.LayerNorm, eps=1e-6)
    else:
        raise NotImplementedError("Only supports layernorm.")
    if isinstance(spatial_size, int):
        spatial_size = (spatial_size, spatial_size)

    conv_patch_op = nn.Conv2d if use_2d_patch else nn.Conv3d
    if enable_patch_embed:
        patch_embed = PatchEmbed2D(
            in_channels=input_channels,
            out_channels=patch_embed_dim,
            conv_kernel_size=conv_patch_embed_kernel,
            conv_stride=conv_patch_embed_stride,
            conv_padding=conv_patch_embed_padding,
            conv=conv_patch_op,
        )
    else:
        patch_embed = None
    input_dims = [temporal_size, spatial_size[0], spatial_size[1]]
    if use_2d_patch:
        input_stirde = (1,) + tuple(conv_patch_embed_stride)
    else:
        input_stirde = conv_patch_embed_stride

    if enable_patch_embed:
        patch_embed_shape = [input_dims[i] // input_stirde[i] for i in range(len(input_dims))]
    else:
        patch_embed_shape = input_dims

    cls_positional_encoding = SpatialTemporalClsPositionalEncoding(
        embed_dim=patch_embed_dim,
        patch_embed_shape=patch_embed_shape,
        sep_pos_embed=sep_pos_embed,
        has_cls=cls_embed_on,
    )

    # stochastic depth decay rule
    dpr = [
        x.item() for x in torch.linspace(0, droppath_rate_block, depth)
    ]

    if dropout_rate_block > 0.0:
        pos_drop = nn.Dropout(p=dropout_rate_block)

    dim_mul, head_mul = torch.ones(depth + 1), torch.ones(depth + 1)
    if embed_dim_mul is not None:
        for i in range(len(embed_dim_mul)):
            dim_mul[embed_dim_mul[i][0]] = embed_dim_mul[i][1]
    if atten_head_mul is not None:
        for i in range(len(atten_head_mul)):
            head_mul[atten_head_mul[i][0]] = atten_head_mul[i][1]

    norm_patch_embed = norm_layer(patch_embed_dim) if enable_patch_embed_norm else None

    pool_q = [[] for i in range(depth)]
    pool_kv = [[] for i in range(depth)]
    stride_q = [[] for i in range(depth)]
    stride_kv = [[] for i in range(depth)]

    if pool_q_stride_size is not None:
        for i in range(len(pool_q_stride_size)):
            stride_q[pool_q_stride_size[i][0]] = pool_q_stride_size[i][1:]
            if pool_kvq_kernel is not None:
                pool_q[pool_q_stride_size[i][0]] = pool_kvq_kernel
            else:
                pool_q[pool_q_stride_size[i][0]] = [
                    s + 1 if s > 1 else s for s in pool_q_stride_size[i][1:]
                ]

    if pool_kv_stride_adaptive is not None:
        stride_kv = pool_kv_stride_adaptive
        pool_kv_stride_size = []
        for i in range(depth):
            if len(stride_q[i]) > 0:
                stride_kv = [
                    max(stride_kv[d] // stride_q[i][d], 1)
                    for d in range(len(stride_kv))
                ]
            pool_kv_stride_size.append([i] + stride_kv)

    if pool_kv_stride_size is not None:
        for i in range(len(pool_kv_stride_size)):
            stride_kv[pool_kv_stride_size[i][0]] = pool_kv_stride_size[i][1:]
            if pool_kvq_kernel is not None:
                pool_kv[pool_kv_stride_size[i][0]] = pool_kvq_kernel
            else:
                pool_kv[pool_kv_stride_size[i][0]] = [
                    s + 1 if s > 1 else s for s in pool_kv_stride_size[i][1:]
                ]

    mvit_blocks = nn.ModuleList()
    for i in range(depth):
        num_heads = round_width(num_heads, head_mul[i], min_width=1, divisor=1)
        patch_embed_dim = round_width(patch_embed_dim, dim_mul[i], divisor=num_heads)
        dim_out = round_width(
            patch_embed_dim,
            dim_mul[i + 1],
            divisor=round_width(num_heads, head_mul[i + 1]),
        )

        mvit_blocks.append(
            MultiScaleBlock(
                dim=patch_embed_dim,
                dim_out=dim_out,
                num_heads=num_heads,
                mlp_ratio=mlp_ratio,
                qkv_bias=qkv_bias,
                dropout_rate=dropout_rate_block,
                droppath_rate=dpr[i],
                norm_layer=norm_layer,
                kernel_q=pool_q[i],
                kernel_kv=pool_kv[i],
                stride_q=stride_q[i],
                stride_kv=stride_kv[i],
                pool_mode=pooling_mode,
                has_cls_embed=cls_embed_on,
                pool_first=pool_first,
            )
        )

    embed_dim = dim_out
    norm_embed = norm_layer(embed_dim)
    if head is not None:
        head_model = head(
            in_features=embed_dim,
            out_features=head_num_classes,
            seq_pool_type="cls" if cls_embed_on else "mean",
            dropout_rate=head_dropout_rate,
            activation=head_activation,
        )
    else:
        head_model = None
    return MultiscaleVisionTransformers(
        patch_embed=patch_embed,
        cls_positional_encoding=cls_positional_encoding,
        pos_drop=pos_drop if dropout_rate_block > 0.0 else None,
        norm_patch_embed=norm_patch_embed,
        blocks=mvit_blocks,
        norm_embed=norm_embed,
        head=head_model,
    )



from torch.utils.data.dataset import Dataset


class TowheeDataSet:
    """
    TowheeDataSet is a kind of dataset wrapper, where the `self.dataset` is the true dataset.
    """
    def __init__(self):
        self.framework = None
        self.dataset = None

    def __len__(self):
        raise NotImplementedError

class TorchDataSet(TowheeDataSet):
    """
    pytorch dataset
    """
    def __init__(self, torchDataSet: Dataset = None):
        super().__init__()
        self.framework = 'torch'
        self.dataset = torchDataSet

    def __len__(self):
        return len(self.dataset)

    def get_framework(self):
        return str(self.framework)



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from towhee.hparam import param_scope
from towhee.operator import Operator


class StatefulOperator(Operator):
    """
    Stateful operator.

    Examples:

    >>> from towhee import register
    >>> from towhee import DataCollection, State
    >>> from towhee.functional.entity import Entity
    >>> import numpy as np
    >>> @register
    ... class my_normalize(StatefulOperator):
    ...     def __init__(self, name):
    ...         super().__init__(name=name)
    ...     def fit(self):
    ...         self._state._mu = np.mean(self._data[0])
    ...         self._state._std = np.std(self._data[0])
    ...     def predict(self, x):
    ...         return (x-self._state._mu)/self._state._std

    >>> dc = (
    ...     DataCollection.range(10)
    ...         .set_training(State())
    ...         .map(lambda x: Entity(a=x))
    ...         .my_normalize['a', 'b'](name='mynorm')
    ... )

    >>> [int(x.b*10) for x in dc.to_list()]
    [-15, -12, -8, -5, -1, 1, 5, 8, 12, 15]
    >>> dc._state.mynorm._mu
    4.5
    """
    def __init__(self, name):
        super().__init__()
        self._name = name if name else self._get_default_name()
        self._data = None
        self._training = False

    def _get_default_name(self):
        with param_scope() as hp:
            name = self.__class__.__name__.replace('_', '-')
            inputs = (hp.index[0] if isinstance(hp.index[0], str) else '-'.join(hp.index[0])).replace('_', '-')
            outputs = (hp.index[1] if isinstance(hp.index[1], str) else '-'.join(hp.index[1])).replace('_', '-')

            return '_'.join([name, inputs, outputs])

    def set_state(self, state):
        if not getattr(state, self._name):
            getattr(state, self._name).model = 1
        self._state = getattr(state, self._name)

    def set_training(self, flag):
        self._training = flag

    def feed(self, *arg):
        if self._data is None:
            self._data = [[x] for x in arg]
            return
        for i in range(len(arg)):
            self._data[i].append(arg[i])

    def fit(self, *arg):
        if len(arg) == 0 and self._data:
            return self.fit(*self._data)

    def predict(self, *arg):
        pass

    def __call__(self, *arg):
        if self._training:
            return self.feed(*arg)
        else:
            return self.predict(*arg)


if __name__ == '__main__':
    import doctest
    doctest.testmod(verbose=False)



# Copyright 2021 biubug6 . All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# This code is modified by Zilliz.

#adapted from https://github.com/biubug6/Pytorch_Retinaface
import torch
from torch import nn

def conv_bn(inp: int, oup: int, stride: int = 1, leaky: int = 0) -> nn.Module:
    return nn.Sequential(
        nn.Conv2d(inp, oup, 3, stride, 1, bias=False),
        nn.BatchNorm2d(oup),
        nn.LeakyReLU(negative_slope=leaky, inplace=True)
    )

def conv_bn_no_relu(inp: int, oup: int, stride: int) -> nn.Module:
    return nn.Sequential(
        nn.Conv2d(inp, oup, 3, stride, 1, bias=False),
        nn.BatchNorm2d(oup),
    )

def conv_bn1x1(inp: int, oup: int, stride: int, leaky: float=0) -> nn.Module:
    return nn.Sequential(
        nn.Conv2d(inp, oup, 1, stride, padding=0, bias=False),
        nn.BatchNorm2d(oup),
        nn.LeakyReLU(negative_slope=leaky, inplace=True)
    )

def conv_dw(inp, oup, stride, leaky=0.1):
    return nn.Sequential(
        nn.Conv2d(inp, inp, 3, stride, 1, groups=inp, bias=False),
        nn.BatchNorm2d(inp),
        nn.LeakyReLU(negative_slope= leaky,inplace=True),

        nn.Conv2d(inp, oup, 1, 1, 0, bias=False),
        nn.BatchNorm2d(oup),
        nn.LeakyReLU(negative_slope= leaky,inplace=True),
    )

# Adapted from https://github.com/Hakuyume/chainer-ssd
def decode(loc, priors, variances):
    """
    Decode locations from predictions using priors to undo
    the encoding we did for offset regression at train time.

    Args:
        loc (`torch.FloatTensor`):
            location predictions for loc layers, Shape: [num_priors,4].
        priors (`torch.FloatTensor`):
            Prior boxes in center-offset form. Shape: [num_priors,4].
        variances (`list[float]`):
            Variances of priorboxes.
    Return:
        (`torch.FloatTensor`)
            decoded bounding box predictions.
    """

    boxes = torch.cat((
        priors[:, :2] + loc[:, :2] * variances[0] * priors[:, 2:],
        priors[:, 2:] * torch.exp(loc[:, 2:] * variances[1])), 1)
    boxes[:, :2] -= boxes[:, 2:] / 2
    boxes[:, 2:] += boxes[:, :2]
    return boxes

def decode_landm(pre, priors, variances):
    """
    Decode landm from predictions using priors to undo
    the encoding we did for offset regression at train time.

    Args:
        pre (`torch.FloatTensor`):
            landm predictions for loc layers, Shape: [num_priors,10].
        priors (`torch.FloatTensor`):
            Prior boxes in center-offset form. Shape: [num_priors,4].
        variances (`list[float]`):
            Variances of priorboxes.
    Return:
        (`torch.FloatTensor`)
            decoded landm predictions.
    """
    landms = torch.cat((priors[:, :2] + pre[:, :2] * variances[0] * priors[:, 2:],
                        priors[:, :2] + pre[:, 2:4] * variances[0] * priors[:, 2:],
                        priors[:, :2] + pre[:, 4:6] * variances[0] * priors[:, 2:],
                        priors[:, :2] + pre[:, 6:8] * variances[0] * priors[:, 2:],
                        priors[:, :2] + pre[:, 8:10] * variances[0] * priors[:, 2:],
                        ), dim=1)
    return landms



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import requests
import random
from requests.exceptions import HTTPError
from requests.auth import HTTPBasicAuth


class HubUtils:
    """
    The Hub Utils to deal with the http request. And other staticmethod for hub.

    Args:
        author (`str`):
            The author of the repo.
        repo (`str`):
            The name of the repo.
        root (`str`):
            The root url where the repo located.
    """
    def __init__(self, author: str = None, repo: str = None, root: str = 'https://towhee.io'):
        self._author = author
        self._repo = repo
        self._root = root

    @property
    def author(self):
        return self._author

    @property
    def repo(self):
        return self._repo

    @property
    def root(self):
        return self._root

    def set_author(self, author: str):
        self._author = author

    def set_repo(self, repo: str):
        self._repo = repo

    def get_info(self) -> requests.Response:
        """
        Get a repo.

        Returns:
            (`Response`)
                The information about the repo.

        Raises:
            (`HTTPError`)
                Raise the error in request.
        """
        url = f'{self._root}/api/v1/repos/{self._author}/{self._repo}'
        response = requests.get(url)
        return response

    def get_commits(self, limit: int, page: int, tag: str) -> requests.Response:
        """
        Get a list of all commits of a repo.

        Args:
            limit (`int`):
                Page size of results.
            page (`int`):
                Page number of results.
            tag (`str`):
                The tag name.

        Returns:
            (`Response`)
                A list of all commits of a repo.

        Raises:
            (`HTTPError`)
                Raise error in request.
        """

        url = f'{self._root}/api/v1/repos/{self._author}/{self._repo}/commits?limit={limit}&page={page}&sha={tag}'
        try:
            response = requests.get(url, allow_redirects=True)
            response.raise_for_status()
            return response
        except HTTPError as e:
            raise e

    def get_tree(self, commit: str) -> requests.Response:
        """
        Get a git tree in the current repo at the given commit.

        Args:
            commit (`str`):
                The commit to base current existing files.

        Returns:
            (`Response`)
                A git tree for the repo

        Raises:
            (`HTTPError`)
                Raise error in request.
        """

        url = f'{self._root}/api/v1/repos/{self._author}/{self._repo}/git/trees/{commit}?recursive=1'
        try:
            response = requests.get(url)
            response.raise_for_status()
            return response
        except HTTPError as e:
            raise e

    def get_file(self, file_path: str, tag: str) -> requests.Response:
        """
        Get a file from a repo.

        Args:
            file_path (`str`):
                filepath of the file to get.
            tag (`str`):
                The tag name.

        Returns:
            (`Response`)
                A file content in the file_path.

        Raises:
            (`HTTPError`)
                Raise error in request.
        """
        url = f'{self._root}/api/v1/repos/{self._author}/{self._repo}/raw/{file_path}?ref={tag}'
        response = requests.get(url, stream=True)
        return response

    def get_lfs(self, file_path: str, tag: str) -> requests.Response:
        """
        Get a file from a repo.

        Args:
            file_path (`str`):
                filepath of the file to get.
            tag (`str`):
                The tag name.

        Returns:
            (`Response`)
                A file content in the file_path.

        Raises:
            (`HTTPError`)
                Raise error in request.
        """
        url = f'{self._root}/{self._author}/{self._repo}/media/branch/{tag}/{file_path}'
        response = requests.get(url, stream=True)
        return response

    def create_token(self, token_name: str, password: str) -> requests.Response:
        """
        Create an account verification token. This token allows for avoiding HttpBasicAuth for subsequent calls.

        Args:
            token_name (`str`):
                The name to be given to the token.
            password (`str`):
                The password of current author.

        Returns:
            (`Response`)
                Return the token(id, name, sha1, token_last_eight).

        Raises:
            (`HTTPError`)
                Raise the error in request.
        """
        url = f'{self._root}/api/v1/users/{self._author}/tokens'
        data = {'name': token_name}
        try:
            response = requests.post(url, data=data, auth=HTTPBasicAuth(self._author, password))
            response.raise_for_status()
            return response
        except HTTPError as e:
            raise e

    def delete_token(self, token_id: int, password: str) -> None:
        """
        Delete the token with the given name. Useful for cleanup after changes.

        Args:
            token_id (`int`):
                The token id.
            password (`str`):
                The password of current author.

        Raises:
            (`HTTPError`)
                Raise error in request.
        """
        url = f'{self._root}/api/v1/users/{self._author}/tokens/{token_id}'
        try:
            response = requests.delete(url, auth=HTTPBasicAuth(self._author, password))
            response.raise_for_status()
        except HTTPError as e:
            raise e

    def create(self, password: str, repo_class: int) -> None:
        """
        Create a repo under current account.

        Args:
            password (`str`):
                Current author's password.
            repo_class (`int`):
                The repo class

        Raises:
            (`HTTPError`)
                Raise error in request.
        """
        token_name = random.randint(0, 10000)
        r = self.create_token(token_name, password)
        token = r.json()

        data = {
            'auto_init': True,
            'default_branch': 'main',
            'description': repo_class,
            'name': self._repo,
            'private': False,
            'template': False,
            'trust_model': 'default',
            'type': repo_class
        }
        url = self._root + '/api/v1/user/repos'
        try:
            r = requests.post(url, data=data, headers={'Authorization': 'token ' + str(token['sha1'])})
            r.raise_for_status()
        except HTTPError as e:
            raise e
        finally:
            self.delete_token(str(token['id']), password)

    def create_repo(self, token: str, repo_class: int) -> None:
        """
        Create a repo under current account with token.

        Args:
            token (`str`):
                Authenticated user's token.

        Raises:
            (`HTTPError`)
                Raise error in request.
        """

        data = {
            'auto_init': True,
            'default_branch': 'main',
            'description': repo_class,
            'name': self._repo,
            'private': False,
            'template': False,
            'trust_model': 'default',
            'type': repo_class
        }
        url = self._root + '/api/v1/user/repos'
        try:
            r = requests.post(url, data=data, headers={'Authorization': 'token ' + str(token)})
            r.raise_for_status()
        except HTTPError as e:
            raise e

    def add_tag(self, password, tag, tag_num):
        """
        Create a repo under current account.

        Args:
            password (`str`):
                Current author's password.
            tag (`str`):
                The tag name to attach.
            tag_num (`int`):
                Which enum # for tag.

        Raises:
            (`HTTPError`)
                Raise error in request.
        """
        token_name = random.randint(0, 10000)
        r = self.create_token(token_name, password)
        token = r.json()
        data = {
            'name': tag,
            'pretty_name': tag,
            'type': tag_num
        }
        url = self._root + '/api/v1/repos/' + self._author + '/' + self._repo + '/repotag'

        try:
            r = requests.post(url, data=data, headers={'Authorization': 'token ' + str(token['sha1'])})
            r.raise_for_status()
        except HTTPError as e:
            raise e
        finally:
            self.delete_token(str(token['id']), password)

    def get_user(self, token: str) -> requests.Response:
        """
        Get the user info with token.

        Args:
            token (`str`):
                Authenticated user's token.

        Returns:
            (`Response`)
                Return the user(id, username etc.).

        Raises:
            (`HTTPError`)
                Raise error in request.
        """
        url = self._root + '/api/v1/user'
        try:
            response = requests.get(url, params={'token': token})
            response.raise_for_status()
            return response
        except HTTPError as e:
            raise e

    def login(self, password, token):
        """
        Login with Authenticated user.

        Args:
            password (`str`):
                Authenticated user's password.
            token (`str`):
                Authenticated user's token.

        Raises:
            (`HTTPError`)
                Raise error in request.
        """
        url = self._root + '/user/login'
        data = {
            'user_name': self._author,
            'password': password,
            '_csrf': token
        }
        try:
            response = requests.post(url, data=data)
            response.raise_for_status()
        except HTTPError as e:
            raise e

    def logout(self, token):
        """
        Logout with token.

        Raises:
            (`HTTPError`)
                Raise error in request.
        """
        url = self._root + '/user/logout'
        data = {'_csrf': token}
        try:
            response = requests.post(url, data=data)
            response.raise_for_status()
        except HTTPError as e:
            raise e
        pass

    @staticmethod
    def convert_dict(dicts: dict) -> dict:
        """
        Convert all the values in a dictionary to str and replace char.

        For example:
        <class 'torch.Tensor'>(unknow type) to torch.Tensor(str type).

        Args:
            dicts (`dict`):
                The dictionary to convert.

        Returns:
            (`dict`)
                The converted dictionary.
        """
        for keys in dicts:
            dicts[keys] = str(dicts[keys]).replace('<class ', '').replace('>', '').replace('\'', '')
        return dict(dicts)

    @staticmethod
    def update_text(ori_str_list: list, tar_str_list: list, file: str, new_file: str) -> None:
        """
        Update the file from ori_str_list to tar_str_list and rename it with the new filename.

        Args:
            ori_str_list (`list`):
                The original str list to be replaced
            tar_str_list (`list`):
                The target str list after replace
            file (`str`):
                The original file name to be updated
            new_file (`str`):
                The target file name after update
        """
        with open(file, 'r', encoding='utf-8') as f1:
            file_text = f1.read()
        # Replace the target string
        for ori_str, tar_str in zip(ori_str_list, tar_str_list):
            file_text = file_text.replace(ori_str, tar_str)
        with open(new_file, 'w', encoding='utf-8') as f2:
            f2.write(file_text)



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class Status:
    """
    Return status
    """

    def __init__(self, code: int, msg: str = None, data: any = None):
        self._code = code
        self._msg = msg
        self._data = data

    @staticmethod
    def ok_status(data: any = None) -> 'Status':
        return Status(0, data=data)

    @staticmethod
    def err_status(msg: str) -> 'Status':
        return Status(-1, msg=msg)

    @property
    def msg(self):
        return self._msg

    @property
    def data(self):
        return self._data

    def is_ok(self):
        return self._code == 0



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import subprocess
from pathlib import Path
from typing import Union, List
from requests.exceptions import HTTPError

from towhee.utils.hub_utils import HubUtils
from towhee.utils.log import engine_log


class GitUtils:
    """
    A wrapper class to wrap git manipulations.

    Args:
        author (`str`):
            The author of the repo.
        repo (`str`):
            The name of the repo.
        root (`str`):
            The root url of the repo.
    """
    def __init__(self, author: str, repo: str, root: str = 'https://towhee.io'):
        self._author = author
        self._repo = repo
        self._root = root

    @property
    def author(self):
        return self._author

    @property
    def repo(self):
        return self._repo

    @property
    def root(self):
        return self._root

    def exists(self) -> bool:
        """
        Check if a repo exists.

        Returns:
            (`bool`)
                return `True` if the repository exists, else `False`.

        Raises:
            (`HTTPError`)
                Raise the error in request.
        """
        try:
            response = HubUtils(self._author, self._repo).get_info()
            return response.status_code == 200
        except HTTPError as e:
            raise e

    def clone(self, tag: str = 'main', install_reqs: bool = True, local_repo_path: Union[str, Path] = None) -> None:
        """
        Clone the repo to specified location.

        Args:
            tag (`str`):
                The tag name.
            install_reqs (`bool`):
                Whether to install packages from requirements.txt.
            self.local_repo_path (`Union[str, Path]`):
                The path to the local repo.

        Raises:
            (`ValueError`)
                Raise error if the repo does not exist.
        """
        if not local_repo_path:
            local_repo_path = Path.cwd() / self._repo_file_name
        local_repo_path = Path(local_repo_path)

        if not self.exists():
            engine_log.error('%s/%s repo does not exist.', self._author, self._repo)
            raise ValueError(f'{self._author}/{self._repo} repo does not exist.')

        url = f'{self._root}/{self._author}/{self._repo}.git'
        try:
            subprocess.check_call(['git', 'lfs'], stdout=subprocess.DEVNULL, stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as e:
            engine_log.warning(
                '\'git-lfs\' not found, execute download instead of clone. ' \
                'If you want to download large file with git-lfs, please install \'git-lfs\' and remove current local cache.'
            )
            raise FileNotFoundError('\'git-lfs\' not found, execute download instead of clone. ' \
                'If you want to download large file with git-lfs, please install \'git-lfs\' and remove current local cache.'
            ) from e

        try:
            subprocess.check_call(['git', 'clone', '-b', tag, url, local_repo_path])
        except FileNotFoundError as e:
            engine_log.warning(
                '\'git\' not found, execute download instead of clone. ' \
                'If you want to check updates every time you run the pipeline, please install \'git\' and remove current local cache.'
            )
            raise FileNotFoundError(
                '\'git\' not found, execute download instead of clone. ' \
                'If you want to check updates every time you run the pipeline, please install \'git\' and remove current local cache.'
            ) from e

        if install_reqs:
            if 'requirements.txt' in os.listdir(local_repo_path):
                subprocess.check_call([sys.executable, '-m', 'pip', 'install', '-r', local_repo_path / 'requirements.txt'])

    def status(self):
        """
        Check the status of local repo.
        """
        try:
            status = subprocess.check_output(['git', 'status'], stderr=subprocess.DEVNULL).decode('utf-8')
            return status
        except subprocess.CalledProcessError as e:
            raise FileNotFoundError('The repo file has not .git.') from e

    def add(self, files: Union[str, List[str]] = None):
        """
        A wrapper function for git add.

        Stage current changes in the repo, please make sure your current woring directory is in the repo.

        Args:
            files (`Union[str, List[str]]`):
                The relative paths of the files to stage regard to the repo.
        """
        if not files:
            return subprocess.check_call(['git', 'add', '.'])

        if isinstance(files, str):
            return subprocess.check_call(['git', 'add', files])

        return subprocess.check_call(['git', 'add'] + files)

    def commit(self, cmt_msg: str):
        """
        A wrapper function for git commit.

        Commit current changes in the repo.

        Args:
            cmt_msg (`str`):
                The commit message.
        """
        return subprocess.check_call(['git', 'commit', '-sm', cmt_msg])

    def push(self, remote: str = 'origin', branch: str = 'main'):
        """
        A wrapper function for git push.

        Push local commits to remote, please make sure your current woring directory is in the repo.

        Args:
            self._repo_path (`Union[str, Path]`):
                The local repo cloned from remote.
            remote (`str`):
                The remote repo.
            branch (`str`):
                The remote branch.
        """
        return subprocess.check_call(['git', 'push', remote, branch])

    def pull(self, remote: str = 'origin', branch: str = 'main'):
        """
        A wrapper function for git pull.

        pull from remote, please make sure your current woring directory is in the repo.

        Args:
            self._repo_path (`Union[str, Path]`):
                The local repo cloned from remote.
            remote (`str`):
                The remote repo.
            branch (`str`):
                The remote branch.
        """
        return subprocess.check_call(['git', 'pull', remote, branch])



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
from getpass import getpass
from pathlib import Path
from shutil import rmtree

from towhee.hub.operator_manager import OperatorManager
from towhee.hub.pipeline_manager import PipelineManager
from towhee.hub.repo_manager import RepoManager
from towhee.utils.git_utils import GitUtils


def get_repo_obj(args):
    if args.command == 'download':
        repo_obj = RepoManager(args.author, args.repo)
    elif args.command == 'clone':
        repo_obj = GitUtils(args.author, args.repo)
    elif args.command == 'generate-yaml' or args.classes in ['pyoperator', 'nnoperator', 'operator']:
        repo_obj = OperatorManager(args.author, args.repo)
    elif args.classes == 'pipeline':
        repo_obj = PipelineManager(args.author, args.repo)
    return repo_obj


def init_repo(args):
    repo_manager = get_repo_obj(args)
    repo_path = Path(args.dir) / args.repo.replace('-', '_')
    GitUtils(args.author, args.repo).clone(local_repo_path=repo_path)
    if args.classes == 'pipeline':
        temp_path = Path(args.dir) / 'pipeline_template'
        GitUtils('towhee', 'pipeline-template').clone(local_repo_path=temp_path)
        repo_manager.init_pipeline(temp_path, repo_path)
    elif args.classes == 'pyoperator':
        temp_path = Path(args.dir) / 'pyoperator_template'
        GitUtils('towhee', 'pyoperator-template').clone(local_repo_path=temp_path)
        repo_manager.init_pyoperator(temp_path, repo_path)
    elif args.classes == 'nnoperator':
        temp_path = Path(args.dir) / 'nnoperator_template'
        GitUtils('towhee', 'nnoperator-template').clone(local_repo_path=temp_path)
        repo_manager.init_nnoperator(temp_path, repo_path, args.framework)
    rmtree(temp_path)


# Get more usage in README.md
def main():
    parser = argparse.ArgumentParser(prog='towhee')
    ar_parser = argparse.ArgumentParser()
    ar_parser.add_argument('-a', '--author', required=True, help='Author of the Repo.')
    ar_parser.add_argument('-r', '--repo', required=True, help='Repo name.')
    c1_parser = argparse.ArgumentParser(add_help=False)
    c1_parser.add_argument('classes', choices=['operator', 'pipeline'], help='Repo class in [\'operator\', \'pipeline\'].')
    c2_parser = argparse.ArgumentParser(add_help=False)
    c2_parser.add_argument(
        '-c', '--classes', choices=['pyoperator', 'nnoperator', 'pipeline'], help='Repo class in [\'pyoperator\', \'nnoperator\', \'pipeline\'].'
    )
    f_parser = argparse.ArgumentParser(add_help=False)
    f_parser.add_argument('-f', '--framework', default='pytorch', help='The framework of nnoperator, defaults to \'pytorch\'.')
    t_parser = argparse.ArgumentParser(add_help=False)
    t_parser.add_argument('-t', '--tag', default='main', help='Repo tag, defaults to \'main\'.')
    d_parser = argparse.ArgumentParser(add_help=False)
    d_parser.add_argument('-d', '--dir', default='.', help='Directory to the Repo file, defaults to \'.\'.')

    subparsers = parser.add_subparsers(dest='command')
    create = subparsers.add_parser('create', parents=[c1_parser, ar_parser], add_help=False, description='Create Repo on Towhee hub.')
    create.add_argument('-p', '--password', nargs='?', required=True, help='Password of the author.')
    subparsers.add_parser(
        'init', parents=[ar_parser, c2_parser, d_parser, f_parser], add_help=False, description='Initialize the file for your Repo.'
    )
    subparsers.add_parser('generate-yaml', parents=[ar_parser, d_parser], add_help=False, description='Generate yaml file for your Operator Repo.')
    subparsers.add_parser('download', parents=[ar_parser, t_parser, d_parser], add_help=False, description='Download repo(without .git) to local.')
    subparsers.add_parser('clone', parents=[ar_parser, t_parser, d_parser], add_help=False, description='Clone repo to local.')

    args = parser.parse_args()
    repo_obj = get_repo_obj(args)

    if args.command == 'create':
        if not args.password:
            args.password = getpass('Password: ')
        print('Creating repo...')
        repo_obj.create(args.password)
        print('Done')
    if args.command == 'init':
        print('Clone the repo and initialize it with template...')
        init_repo(args)
        print('Done')
    elif args.command == 'generate-yaml':
        print('Generating yaml for repo...')
        repo_obj.generate_yaml(Path(args.dir))
        print('Done')
    elif args.command == 'download':
        print('Downloading repo...')
        repo_obj.download(Path(args.dir) / args.repo, args.tag, False)
        print('Done')
    elif args.command == 'clone':
        print('Cloning repo...')
        repo_obj.clone(args.tag, False, Path(args.dir) / args.repo)
        print('Done')


if __name__ == '__main__':
    main()



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import os
import argparse
import subprocess
from pathlib import Path
from setuptools import setup

from typing import List

parser = argparse.ArgumentParser(add_help=False)
parser.add_argument('-n', '--namespace', default='towhee', help='optional, repo author or namespace, defaults to \'towhee\'')
parser.add_argument('-p', '--path', required=False, default='.', help='optional, path to the operator repo, defaults to cwd which is \'.\'')


class SetupCommand:
    """
    Implementation for subcmd `towhee develop` and `towhee install`.
    Setup repo to `towheeoperator.{self._args.namespace}_{self._args.repo_name}' package with pypi methods.
    """

    def __init__(self, args) -> None:
        self._args = args
        self._path = Path.home() / '.towhee' / 'towheeoperator'

    def __call__(self) -> None:
        path = Path(self._args.path).resolve()
        repo_name = path.stem.replace('-', '_')
        package_name = f'towheeoperator.{self._args.namespace}_{repo_name}'

        packages = [package_name]
        for p in path.iterdir():
            if p.is_dir() and not p.name.startswith('.') and not p.name.endswith('.egg-info') and p.name not in ['__pycache__', 'build', 'dist']:
                packages.append(f'{package_name}.{p.name}')

        requirements = self.read_requirements()
        if self._args.action == 'install':
            if not self._args.develop:
                os.chdir(str(path))
                if not self._args.user:
                    sys.argv = ['setup.py', 'install']
                else:
                    sys.argv = ['setup.py', 'install', '--user']
            else:
                pypi_path = self._path / repo_name
                symlink_path = pypi_path / 'towheeoperator'
                if not symlink_path.exists():
                    symlink_path.mkdir(parents=True)

                symlink = symlink_path / f'{self._args.namespace}_{repo_name}'
                symlink.unlink(missing_ok=True)
                packages = [package_name]
                symlink.symlink_to(path)
                os.chdir(str(pypi_path))
                if not self._args.user:
                    sys.argv = ['setup.py', 'develop']
                else:
                    sys.argv = ['setup.py', 'develop', '--user']

        setup(name=package_name,
              packages=packages,
              package_dir={package_name: '.'},
              package_data={'': ['*.txt', '*.rst']},
              install_requires=requirements,
              zip_safe=False)

    @staticmethod
    def read_requirements() -> List[str]:
        """
        Get a list in requirements.txt.

        Return:
            (`List`):
                List in ./requirements.txt.
        """
        try:
            with open('requirements.txt', encoding='utf-8') as f:
                required = f.read().splitlines()
        except FileNotFoundError:
            required = []
        return required

    @staticmethod
    def install(subparsers):
        install = subparsers.add_parser('install', parents=[parser], help='setup command: install operator with setup.py')
        install.add_argument('--develop', action='store_true', help='optional, install operator with setup.py develop')
        install.add_argument('--user', action='store_true', help='optional, whether to install op to the local path(site.USER_SITE).')


class UninstallCommand:
    """
    Implementation for subcmd `towhee uninstall`.
    """
    def __init__(self, args) -> None:
        self._args = args

    def __call__(self) -> None:
        path = Path(self._args.path).resolve()
        repo_name = path.stem.replace('-', '_')
        subprocess.check_call([sys.executable, '-m', 'pip', 'uninstall', '-y', f'towheeoperator.{self._args.namespace}_{repo_name}'])

    @staticmethod
    def install(subparsers):
        subparsers.add_parser('uninstall', parents=[parser], help='setup command: uninstall operator with pip uninstall')



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import threading
from typing import Tuple

from towhee.dataframe import DataFrame
from towhee.dag import GraphRepr
from towhee.dataframe.iterators import MapIterator

from towhee.engine.operator_context import OperatorContext, OpStatus
from towhee.utils.log import engine_log
from towhee.errors import OpFailedError


class GraphContext:
    """
    `GraphContext` is a data processing network with one or multiple `Operator`.
    Each row of a `Pipeline`'s inputs will be processed individually by a
    `GraphContext`.

    Args:
        ctx_idx: (`int`)
            The index of this `GraphContext`.
        graph_repr: (`towhee.dag.GraphRepr`)
            The DAG representation this `GraphContext` will implement.
    """

    def __init__(self, ctx_idx: int, graph_repr: GraphRepr):
        self._idx = ctx_idx
        self._repr = graph_repr

        self._lock = threading.Lock()
        self._build_components()

    def __call__(self, inputs: Tuple):
        self.inputs.put(inputs)
        self.inputs.seal()

    @property
    def inputs(self) -> DataFrame:
        """Returns the graph's input `DataFrame`.

        Returns:
            (`towhee.dataframe.DataFrame`)
                The input `DataFrame`.
        """
        return self.dataframes['_start_df']

    @property
    def outputs(self) -> DataFrame:
        """Returns the graph's output `DataFrame`.

        Returns:
            (`towhee.dataframe.DataFrame`)
                The output `DataFrame`.
        """
        return self.dataframes['_end_df']

    def result(self) -> any:
        if self.outputs.size != 0:
            return self._out_iter
        else:
            # graph run failed, raise an exception
            for op in self._op_ctxs.values():
                if op.status == OpStatus.FAILED:
                    raise OpFailedError(op.err_msg)
            engine_log.warning('The pipeline runs successfully, but no data return')
            return None

    @property
    def op_ctxs(self):
        return self._op_ctxs

    @property
    def dataframes(self):
        return self._dataframes

    def slow_down(self, df_name: str, time_sec: int):
        '''
        Slow down the op whose df name it.
        '''
        self._df_op[df_name].slow_down(time_sec)

    def speed_up(self, df_name: str):
        '''
        spped up the op whose df name it.
        '''
        self._df_op[df_name].speed_up()

    def stop(self):
        for op in self._op_ctxs:
            op.stop()

    def gc(self):
        for _, df in self._dataframes.items():
            df.gc()

    def join(self):
        for op in self._op_ctxs.values():
            op.join()

    def _build_components(self):
        """
        Builds `DataFrame`s and `OperatorContext`s required to run this graph.
        """

        # Build dataframes.
        dfs = {}
        for df_name, df_repr in self._repr.dataframes.items():
            cols = [(col.name, col.vtype) for col in df_repr.columns]
            dfs[df_name] = DataFrame(df_name, cols)
        self._dataframes = dfs

        # Build operator contexts.
        self._op_ctxs = {}
        self._df_op = {}
        for _, op_repr in self._repr.operators.items():
            op_ctx = OperatorContext(op_repr, self.dataframes)
            self._df_op[op_repr.outputs[0]['df']] = op_ctx
            self._op_ctxs[op_ctx.name] = op_ctx
        self._out_iter = MapIterator(self.outputs)

    def __del__(self):
        engine_log.info('Graph % s end', self._idx)



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random
import sys
import subprocess
from getpass import getpass
from requests.exceptions import HTTPError

from towhee.utils.hub_utils import HubUtils
from towhee.utils.hub_file_utils import HubFileUtils


class UserCommand:
    """
    Implementation for subcmd `towhee login`, `towhee whoami` and `towhee logout`.
    """
    def __init__(self, args) -> None:
        self._args = args
        self._root = 'https://towhee.io'
        self.hub = HubUtils()
        self.file = HubFileUtils()
        self._token = self.file.token

    def __call__(self) -> None:
        if self._args.action == 'login':
            self.login()
        elif self._args.action == 'whoami':
            username = self.whoami()
            if username:
                print(f'Username: {username}')
        elif self._args.action == 'logout':
            self.logout()

    @staticmethod
    def install(subparsers):
        subparsers.add_parser('login', help='user command: login using the same credentials as on towhee.io')

    def login(self) -> None:
        """
        Log in with https://towhee.io account.
        """
        if not self._token:
            username = input('Username: ')
            password = getpass()
            self.hub.set_author(username)
            try:
                r = self.hub.create_token(random.randint(0, 10000), password)
                token = r.json()['sha1']
            except HTTPError:
                print('Error password.')
                sys.exit()
            self.hub.login(password, token)
            self.write_to_credential_store(username, token)
            self.file.set_token(token)
            self.file.save()
            print('Successfully logged in.')
        else:
            print('You are already logged in, please log out first.')

    def whoami(self) -> str:
        if self._token:
            res = self.hub.get_user(self._token).json()
            username = res['username']
            return username
        else:
            print('Not logged it.')

    def logout(self) -> None:
        """
        Log out with https://towhee.io account.
        """
        if self._token:
            username = self.whoami()
            self.hub.set_author(username)
            self.hub.logout(self._token)
            self.erase_from_credential_store(username)
            self.file.delete()
            print('Done.')
        else:
            print('Not logged it.')

    def write_to_credential_store(self, username: str, password: str) -> None:
        """
        Write a token to store credentials.

        Args:
            username (`str`):
                authorized username.
            password (`str`):
                user password.
        """
        with subprocess.Popen(
            'git credential-store store'.split(),
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        ) as process:
            input_username = f'username={username.lower()}'
            input_password = f'password={password}'

            process.stdin.write(
                f'url={self._root}\n{input_username}\n{input_password}\n\n'.encode('utf-8')
            )
            process.stdin.flush()

    def erase_from_credential_store(self, username=None) -> None:
        """
        Erases the credential store relative to https://towhee.io. If no `username` is specified, will erase the first entry.

        Args:
            username (`str`):
                authorized username.
        """
        with subprocess.Popen(
                'git credential-store erase'.split(),
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
        ) as process:
            standard_input = f'url={self._root}\n'

            if username is not None:
                standard_input += f'username={username.lower()}\n'

            standard_input += '\n'

            process.stdin.write(standard_input.encode('utf-8'))
            process.stdin.flush()


class LogoutCommand:
    def __init__(self, args) -> None:
        self._args = args

    def __call__(self) -> None:
        UserCommand(self._args)()

    @staticmethod
    def install(subparsers):
        subparsers.add_parser('logout', help='user command: logout')


class WhoCommand:
    def __init__(self, args) -> None:
        self._args = args

    def __call__(self) -> None:
        UserCommand(self._args)()

    @staticmethod
    def install(subparsers):
        subparsers.add_parser('whoami', help='user command: find out which towhee.io account you are logged in')



# Copyright 2021  Facebook. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# This code is modified by Zilliz.


from torch import nn

from towhee.models.layers.resnet_basic_3d_module import ResNetBasic3DModule


def create_resnet_basic_3d_module(
    *,
    in_channels,
    out_channels,
    conv_kernel_size=(3, 7, 7),
    conv_stride=(1, 2, 2),
    conv_padding=(1, 3, 3),
    conv_bias=False,
    conv=nn.Conv3d,
    pool=nn.MaxPool3d,
    pool_kernel_size=(1, 3, 3),
    pool_stride=(1, 2, 2),
    pool_padding=(0, 1, 1),
    norm=nn.BatchNorm3d,
    norm_eps=1e-5,
    norm_momentum=0.1,
    activation=nn.ReLU,
) -> nn.Module:
    """
    ResNet basic 3D stem module. It performs spatiotemporal Convolution, BN, and activation
    following by a spatiotemporal pooling.
                                        Conv3d
                                           ↓
                                     Normalization
                                           ↓
                                       Activation
                                           ↓
                                        Pool3d
    Args:
        in_channels (`int`):
            Input channel size of the convolution.
        out_channels (`int`):
            Output channel size of the convolution.
        conv_kernel_size (`Tuple`):
            Convolutional kernel size(s).
        conv_stride (`Tuple`):
            Convolutional stride size(s).
        conv_padding (`Tuple`):
            Convolutional padding size(s).
        conv_bias (`bool`):
            Convolutional bias. If true, adds a learnable bias to the output.
        conv (`callable`):
            Callable used to build the convolution layer.
        pool (`Callable`):
            A callable that constructs pooling layer, options include: nn.AvgPool3d,
            nn.MaxPool3d, and None (not performing pooling).
        pool_kernel_size (`Tuple`):
            Pooling kernel size(s).
        pool_stride (`Tuple`):
            Pooling stride size(s).
        pool_padding (`Tuple`):
            Pooling padding size(s).
        norm (`Callable`):
            A callable that constructs normalization layer, options
            include nn.BatchNorm3d, None (not performing normalization).
        norm_eps (`float`):
            Normalization epsilon.
        norm_momentum (`float`):
            Normalization momentum.
        activation (`Callable`):
            A callable that constructs activation layer, options
            include: nn.ReLU, nn.Softmax, nn.Sigmoid, and None (not performing activation).
    """

    assert conv is not None

    conv_module = conv(
        in_channels=in_channels,
        out_channels=out_channels,
        kernel_size=conv_kernel_size,
        stride=conv_stride,
        padding=conv_padding,
        bias=conv_bias,
    )
    norm_module = (
        None
        if norm is None
        else norm(num_features=out_channels, eps=norm_eps, momentum=norm_momentum)
    )
    activation_module = (
        None
        if activation is None
        else activation()
    )
    pool_module = (
        None
        if pool is None
        else pool(
            kernel_size=pool_kernel_size, stride=pool_stride, padding=pool_padding
        )
    )
    return ResNetBasic3DModule(
        conv=conv_module,
        norm=norm_module,
        activation=activation_module,
        pool=pool_module,
        )



import torch
import torch.nn.functional as F
from torch import nn


class ASPP(nn.Module):
    """
    Atrous spatial pyramid pooling used in object detection and segmentation.
    """
    def __init__(self, in_channel=512, depth=256):
        super().__init__()
        # global average pooling : init nn.AdaptiveAvgPool2d ;also forward torch.mean(,,keep_dim=True)
        self.mean = nn.AdaptiveAvgPool2d((1, 1))
        self.conv = nn.Conv2d(in_channel, depth, 1, 1)
        # k=1 s=1 no pad
        self.atrous_block1 = nn.Conv2d(in_channel, depth, 1, 1)
        self.atrous_block6 = nn.Conv2d(in_channel, depth, 3, 1, padding=6, dilation=6)
        self.atrous_block12 = nn.Conv2d(in_channel, depth, 3, 1, padding=12, dilation=12)
        self.atrous_block18 = nn.Conv2d(in_channel, depth, 3, 1, padding=18, dilation=18)

        self.conv_1x1_output = nn.Conv2d(depth * 5, depth, 1, 1)

    def forward(self, x):
        size = x.shape[2:]

        image_features = self.mean(x)
        image_features = self.conv(image_features)
        image_features = F.upsample(image_features, size=size, mode='bilinear')

        atrous_block1 = self.atrous_block1(x)

        atrous_block6 = self.atrous_block6(x)

        atrous_block12 = self.atrous_block12(x)

        atrous_block18 = self.atrous_block18(x)

        net = self.conv_1x1_output(torch.cat([image_features, atrous_block1, atrous_block6,
                                              atrous_block12, atrous_block18], dim=1))
        return net



# Copyright 2021 Ross Wightman . All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# This code is modified by Zilliz.

import torch
from torch import nn
from torch.nn import functional as F

def hard_swish(x: torch.Tensor, inplace: bool = False) -> torch.Tensor:
    inner = F.relu6(x + 3.).div_(6.)
    return x.mul_(inner) if inplace else x.mul(inner)

class HardSwish(nn.Module):
    """
    HardSwish activiation layer.

    Applies the hardswish function, element-wise.
    Described in: https://arxiv.org/abs/1905.02244.

    Args:
        inplace(`Bool`):
            whether use inplace version.
    Returns:
        (`torch.Tensor`)
            output tensor after activation.
    """
    def __init__(self, inplace: bool = False) -> None:
        super().__init__()
        self.inplace = inplace

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return hard_swish(x, self.inplace)



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from queue import Queue

from towhee.engine.operator_io.reader import ReaderBase

class StopFrame:
    pass


class MockReader(ReaderBase):
    """
    MockReader for engine test
    """
    def __init__(self, queue: Queue):
        self._queue = queue

    def read(self):
        # Blocking if queue is empty
        data = self._queue.get()

        if not isinstance(data, StopFrame):
            return data
        else:
            raise StopIteration()

    def close(self):
        self._queue.put(StopFrame())



# original code from https://github.com/SwinTransformer/Video-Swin-Transformer
# modified by Zilliz.


def get_window_size(x_size, window_size, shift_size=None):
    """
    Args:
        x_size (`tuple[int]`):
            Tensor with size (B*num_windows, window_size, window_size, C)
        window_size (`tuple[int]`):
            3D window size.
        shift_size (`tuple[int]`):
            Shift size. ref: https://arxiv.org/pdf/2106.13230.pdf
    Returns (`tuple[int]`):
        shifted window size.
    """
    use_window_size = list(window_size)
    if shift_size is not None:
        use_shift_size = list(shift_size)
    for i in range(len(x_size)):
        if x_size[i] <= window_size[i]:
            use_window_size[i] = x_size[i]
            if shift_size is not None:
                use_shift_size[i] = 0

    if shift_size is None:
        return tuple(use_window_size)
    else:
        return tuple(use_window_size), tuple(use_shift_size)



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any, Dict

import torchmetrics as tm
from torchmetrics.detection.map import MeanAveragePrecision

__all__ = [
    'Metrics',
    'TMMetrics',
    'show_avaliable_metrics',
    'get_metric_by_name'
]

_metrics_impls = {}
_metrics_meta = {}

class Metrics:
    """
    `Metrics` is wrapper of popular metrics implementation used to evaluate the
    current prediction's performance by comparing it with the labels.

    Args:
        metric_name (`str`):
            Indicate which metric is used to evalute.
    """

    @classmethod
    def _class_register(cls):
        Metrics._subclass_registry[cls.__repr__] = cls

    def __init__(self, metric_name: str):
        pass

    def update(self, *_: Any, **__: Any) -> None:
        """
        Update the metric internal status.
        """
        raise NotImplementedError

    def compute(self) -> float:
        """
        Get the metric value.
        """
        raise NotImplementedError

    def reset(self) -> None:
        """
        Reset the internal status of 'Metrics'.
        """
        raise NotImplementedError

    def to(self, *args, **kwargs):
        """
        Make 'Metrics' calculation on specific device.
        """
        raise NotImplementedError

class TMMetrics(Metrics):
    """
    `TMMetrics` use torchmetric as the implementation of metrics.

    Args:
        metric_name (`str`):
            Indicate which metric is used to evalute.

    Example:
        >>> from towhee.trainer.metrics import show_avaliable_metrics, get_metric_by_name
        >>> from towhee.trainer.metrics import TMMetrics
        >>> import torch
        >>> pred = torch.FloatTensor([1.0,1.0,1.0,0.0,0.0,0.0])
        >>> gt = torch.IntTensor([1,1,1,1,1,1])
        >>> metric =  TMMetrics('Accuracy')
        >>> metric.update(pred, gt)
        >>> pred = torch.FloatTensor([1.0,1.0,1.0,1.0,1.0,1.0])
        >>> gt = torch.IntTensor([1,1,1,1,1,1])
        >>> metric.update(pred, gt)
        >>> metric.compute().item()
        0.75
        >>> metric.reset()
        >>> pred = torch.FloatTensor([1.0,1.0,1.0,0.0,0.0,0.0])
        >>> gt = torch.IntTensor([1,1,1,1,1,1])
        >>> metric.update(pred, gt)
        >>> metric.compute().item()
        0.5
    """
    @classmethod
    def get_tm_avaliable_metrics(cls):
        cls._metrics_list = {}
        cls._metrics_list['CatMetric'] = tm.CatMetric
        cls._metrics_list['MaxMetric'] = tm.MaxMetric
        cls._metrics_list['MeanMetric'] = tm.MeanMetric
        cls._metrics_list['MinMetric'] = tm.MinMetric
        cls._metrics_list['SumMetric'] = tm.SumMetric
        cls._metrics_list['PIT'] = tm.PIT
        cls._metrics_list['SDR'] = tm.SDR
        cls._metrics_list['SI_SDR'] = tm.SI_SDR
        cls._metrics_list['SI_SNR'] = tm.SI_SNR
        cls._metrics_list['SNR'] = tm.SNR
        cls._metrics_list['PermutationInvariantTraining'] = tm.PermutationInvariantTraining
        cls._metrics_list['ScaleInvariantSignalDistortionRatio'] = tm.ScaleInvariantSignalDistortionRatio
        cls._metrics_list['ScaleInvariantSignalNoiseRatio'] = tm.ScaleInvariantSignalNoiseRatio
        cls._metrics_list['SignalDistortionRatio'] = tm.SignalDistortionRatio
        cls._metrics_list['SignalNoiseRatio'] = tm.SignalNoiseRatio
        cls._metrics_list['AUC'] = tm.AUC
        cls._metrics_list['AUROC'] = tm.AUROC
        cls._metrics_list['F1'] = tm.F1
        cls._metrics_list['ROC'] = tm.ROC
        cls._metrics_list['Accuracy'] = tm.Accuracy
        cls._metrics_list['AveragePrecision'] = tm.AveragePrecision
        cls._metrics_list['BinnedAveragePrecision'] = tm.BinnedAveragePrecision
        cls._metrics_list['BinnedPrecisionRecallCurve'] = tm.BinnedPrecisionRecallCurve
        cls._metrics_list['BinnedRecallAtFixedPrecision'] = tm.BinnedRecallAtFixedPrecision
        cls._metrics_list['CalibrationError'] = tm.CalibrationError
        cls._metrics_list['CohenKappa'] = tm.CohenKappa
        cls._metrics_list['ConfusionMatrix'] = tm.ConfusionMatrix
        cls._metrics_list['F1Score'] = tm.F1Score
        cls._metrics_list['FBeta'] = tm.FBeta
        cls._metrics_list['FBetaScore'] = tm.FBetaScore
        cls._metrics_list['HammingDistance'] = tm.HammingDistance
        cls._metrics_list['Hinge'] = tm.Hinge
        cls._metrics_list['HingeLoss'] = tm.HingeLoss
        cls._metrics_list['IoU'] = tm.IoU
        cls._metrics_list['JaccardIndex'] = tm.JaccardIndex
        cls._metrics_list['KLDivergence'] = tm.KLDivergence
        cls._metrics_list['MatthewsCorrcoef'] = tm.MatthewsCorrcoef
        cls._metrics_list['MatthewsCorrCoef'] = tm.MatthewsCorrCoef
        cls._metrics_list['Precision'] = tm.Precision
        cls._metrics_list['PrecisionRecallCurve'] = tm.PrecisionRecallCurve
        cls._metrics_list['Recall'] = tm.Recall
        cls._metrics_list['Specificity'] = tm.Specificity
        cls._metrics_list['PSNR'] = tm.PSNR
        cls._metrics_list['SSIM'] = tm.SSIM
        cls._metrics_list['MultiScaleStructuralSimilarityIndexMeasure'] = tm.MultiScaleStructuralSimilarityIndexMeasure
        cls._metrics_list['PeakSignalNoiseRatio'] = tm.PeakSignalNoiseRatio
        cls._metrics_list['StructuralSimilarityIndexMeasure'] = tm.StructuralSimilarityIndexMeasure
        cls._metrics_list['CosineSimilarity'] = tm.CosineSimilarity
        cls._metrics_list['ExplainedVariance'] = tm.ExplainedVariance
        cls._metrics_list['MeanAbsoluteError'] = tm.MeanAbsoluteError
        cls._metrics_list['MeanAbsolutePercentageError'] = tm.MeanAbsolutePercentageError
        cls._metrics_list['MeanSquaredError'] = tm.MeanSquaredError
        cls._metrics_list['MeanSquaredLogError'] = tm.MeanSquaredLogError
        cls._metrics_list['PearsonCorrcoef'] = tm.PearsonCorrcoef
        cls._metrics_list['PearsonCorrCoef'] = tm.PearsonCorrCoef
        cls._metrics_list['R2Score'] = tm.R2Score
        cls._metrics_list['SpearmanCorrcoef'] = tm.SpearmanCorrcoef
        cls._metrics_list['SpearmanCorrCoef'] = tm.SpearmanCorrCoef
        cls._metrics_list['SymmetricMeanAbsolutePercentageError'] = tm.SymmetricMeanAbsolutePercentageError
        cls._metrics_list['TweedieDevianceScore'] = tm.TweedieDevianceScore
        cls._metrics_list['RetrievalFallOut'] = tm.RetrievalFallOut
        cls._metrics_list['RetrievalHitRate'] = tm.RetrievalHitRate
        cls._metrics_list['RetrievalMAP'] = tm.RetrievalMAP
        cls._metrics_list['RetrievalMRR'] = tm.RetrievalMRR
        cls._metrics_list['RetrievalNormalizedDCG'] = tm.RetrievalNormalizedDCG
        cls._metrics_list['RetrievalPrecision'] = tm.RetrievalPrecision
        cls._metrics_list['RetrievalRecall'] = tm.RetrievalRecall
        cls._metrics_list['RetrievalRPrecision'] = tm.RetrievalRPrecision
        cls._metrics_list['WER'] = tm.WER
        cls._metrics_list['BLEUScore'] = tm.BLEUScore
        cls._metrics_list['CharErrorRate'] = tm.CharErrorRate
        cls._metrics_list['CHRFScore'] = tm.CHRFScore
        cls._metrics_list['ExtendedEditDistance'] = tm.ExtendedEditDistance
        cls._metrics_list['MatchErrorRate'] = tm.MatchErrorRate
        cls._metrics_list['SacreBLEUScore'] = tm.SacreBLEUScore
        cls._metrics_list['SQuAD'] = tm.SQuAD
        cls._metrics_list['TranslationEditRate'] = tm.TranslationEditRate
        cls._metrics_list['WordErrorRate'] = tm.WordErrorRate
        cls._metrics_list['WordInfoLost'] = tm.WordInfoLost
        cls._metrics_list['WordInfoPreserved'] = tm.WordInfoPreserved
        cls._metrics_list['MinMaxMetric'] = tm.MinMaxMetric
        cls._metrics_list['MeanAveragePrecision'] = MeanAveragePrecision

        return list(cls._metrics_list.keys())

    def __init__(self, metric_name: str):
        super().__init__(metric_name)
        assert metric_name in TMMetrics._metrics_list
        self._metric = TMMetrics._metrics_list[metric_name]()

    def update(self, *_: Any, **__: Any) -> None:
        self._metric.update(*_,**__)

    def compute(self) -> str:
        return self._metric.compute()

    def reset(self) -> None:
        self._metric.reset()

    def to(self, *args, **kwargs):
        self._metric.to(*args, **kwargs)


_metrics_impls['TMMetrics'] = TMMetrics

def _generate_meta_info() -> None:
    global _metrics_meta
    res = {}
    res['TMMetrics'] = {}
    res['TMMetrics']['impl'] = 'pytorch'
    res['TMMetrics']['names'] = _metrics_impls['TMMetrics'].get_tm_avaliable_metrics()
    _metrics_meta = res

def show_avaliable_metrics() -> Dict:
    """
    Get the dict of current avaliable metrics.

    Returns:
        (`Dict`)
            A Dict which key is the class name of metrics implementation(e.g. TMMetrics),
            and the value is a dict has keys 'impl' and 'names'. 'impl' is the framework
            to implement this metric class and 'names' is the list of names of metrics
            which can be created to calculate the metric.

    Example:
        >>> from towhee.trainer.metrics import show_avaliable_metrics
        >>> aval_metrics = show_avaliable_metrics()
        >>> 'TMMetrics' in aval_metrics
        True
    """
    global _metrics_meta
    return _metrics_meta

def get_metric_by_name(metric_name: str, metric_impl: str = 'TMMetrics') -> Metrics:
    """
    Get the `Metrics` class by metric names, and metric_impl should be the specific implementation class.

    Args:
        metric_name (`str`):
            The metric name used to evaluate. (e.g. Accuracy)
        metric_impl (`str`):
            The metric implementation class name. (e.g. TMMetrics)

    Returns:
        (`towhee.trainer.metrics.Metrics`)
            Metrics instance which name matches the queried name.

    Example:
        >>> from towhee.trainer.metrics import show_avaliable_metrics, get_metric_by_name
        >>> from towhee.trainer.metrics import TMMetrics
        >>> metric = get_metric_by_name('Accuracy', 'TMMetrics')
        >>> isinstance(metric, TMMetrics)
        True
    """
    global _metrics_impls
    return _metrics_impls[metric_impl](metric_name)

_generate_meta_info()



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from towhee.utils.log import engine_log

try:
    # pylint: disable=unused-import
    from sklearn.metrics import confusion_matrix, accuracy_score, recall_score, ConfusionMatrixDisplay
    from sklearn.model_selection import train_test_split
    from sklearn.linear_model import LogisticRegression
    from sklearn.ensemble import RandomForestClassifier
    from sklearn.tree import DecisionTreeClassifier
    from sklearn import svm
except ModuleNotFoundError as e:
    engine_log.error('sklearn not found, you can install via `pip install scikit-learn`.')
    raise ModuleNotFoundError('sklearn not found, you can install via `pip install scikit-learn`.') from e



# Copyright 2021  Facebook. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# This code is modified by Zilliz.


import math


def round_width(width, multiplier, min_width=8, divisor=8, ceil=False):
    """
    Round width of filters based on width multiplier
    Args:
        width ('int'):
            The channel dimensions of the input.
        multiplier ('float'):
            The multiplication factor.
        min_width ('int'):
            The minimum width after multiplication.
        divisor ('int'):
            The new width should be dividable by divisor.
        ceil ('bool'):
            If True, use ceiling as the rounding method.
    """
    if not multiplier:
        return width

    width *= multiplier
    min_width = min_width or divisor
    if ceil:
        width_out = max(min_width, int(math.ceil(width / divisor)) * divisor)
    else:
        width_out = max(min_width, int(width + divisor / 2) // divisor * divisor)
    if width_out < 0.9 * width:
        width_out += divisor
    return int(width_out)



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np


class AudioFrame(np.ndarray):
    """
    This class represents an audio frame. The data is a numpy.ndarray.

    Args:
        sample_rate (`int`):
            The audio sample rate in Hz.
    """

    def __new__(cls, data: np.ndarray, sample_rate: int = None, timestamp: int = None):
        # Cast `np.ndarray` to be `AudioFrame`.
        # See https://numpy.org/doc/stable/user/basics.subclassing.html for details.
        obj = np.asarray(data).view(cls)
        obj._sample_rate = sample_rate
        obj._timestamp = timestamp
        return obj

    def __array_finalize__(self, obj):
        # `self` is a new object resulting from super().__new__(cls, ...), therefore it
        # only has attributes that the ndarray.__new__ constructor gave it -
        # i.e. those of a standard ndarray.
        if obj is None:
            return
        self._sample_rate = getattr(obj, '_sample_rate', None)
        self._timestamp = getattr(obj, '_timestamp', None)

    def __str__(self):
        return 'AudioFrame' \
            + ' shape=' + repr(self.shape) \
            + ' sample_rate=' + self.sample_rate \
            + ' timestamp=' + self.timestamp

    @property
    def sample_rate(self):
        return self._sample_rate

    @property
    def timestamp(self):
        return self._timestamp
